# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Image and tile manipulation (transforming, merging, etc).
"""
import io
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional, Union, IO, cast, Any

from PIL import Image, ImageChops
from PIL.TiffImagePlugin import ImageFileDirectory_v2, TiffTags
from mapproxy.image.opts import create_image, ImageFormat
from mapproxy.config import base_config
from mapproxy.srs import make_lin_transf, get_epsg_num

import logging
from functools import reduce
log = logging.getLogger('mapproxy.image')


magic_bytes = [
    ('png', (b"\211PNG\r\n\032\n",)),
    ('jpeg', (b"\xFF\xD8",)),
    ('tiff', (b"MM\x00\x2a", b"II\x2a\x00",)),
    ('gif', (b"GIF87a", b"GIF89a",)),
]


def peek_image_format(buf):
    buf.seek(0)
    header = buf.read(10)
    buf.seek(0)
    for format, bytes in magic_bytes:
        if header.startswith(bytes):
            return format
    return None


TIFF_MODELPIXELSCALETAG = 33550
TIFF_MODELTIEPOINTTAG = 33922
TIFF_GEOKEYDIRECTORYTAG = 34735


class GeoReference(object):
    def __init__(self, bbox, srs):
        self.bbox = bbox
        self.srs = srs

    def tiepoints(self):
        return (
            0.0, 0.0, 0.0,
            self.bbox[0], self.bbox[3], 0.0,
        )

    def pixelscale(self, img_size):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        return (
            float(width)/img_size[0], float(height)/img_size[1], 0.0,
        )

    def tiff_tags(self, img_size):
        tags = ImageFileDirectory_v2()
        tags[TIFF_MODELPIXELSCALETAG] = self.pixelscale(img_size)
        tags.tagtype[TIFF_MODELPIXELSCALETAG] = TiffTags.DOUBLE
        tags[TIFF_MODELTIEPOINTTAG] = self.tiepoints()
        tags.tagtype[TIFF_MODELTIEPOINTTAG] = TiffTags.DOUBLE

        model_type = 2 if self.srs.is_latlong else 1
        tags[TIFF_GEOKEYDIRECTORYTAG] = (
            1, 1, 0, 3,  # {KeyDirectoryVersion, KeyRevision, MinorRevision, NumberOfKeys}
            1024, 0, 1, model_type,  # 1 projected, 2 geographic (lat/long)
            1025, 0, 1, 1,  # 1 RasterIsArea, 2 RasterIsPoint
            3072, 0, 1, get_epsg_num(self.srs.srs_code),
        )
        tags.tagtype[TIFF_GEOKEYDIRECTORYTAG] = TiffTags.SHORT
        return tags


class BaseImageResult(ABC):
    """
    Virtual parent class for ImageResult and BlankImageResult
    """

    size: tuple[int, int]
    image_opts: Any
    cacheable: bool
    authorize_stale: bool

    @abstractmethod
    def as_image(self) -> Image.Image:
        pass

    @abstractmethod
    def as_buffer(self, image_opts=None, format: Optional[str] = None, seekable: bool = False) -> \
            Union[IO[bytes], 'ReadBufWrapper']:
        pass

    def close_buffers(self):
        pass


class ImageResult(BaseImageResult):
    """
    This class wraps either a PIL image, a file-like object, or a file name.
    You can access the result as an image (`as_image` ) or a file-like buffer
    object (`as_buffer`).
    """

    def __init__(self, image: Union[Image.Image, IO[bytes], str], size: Optional[tuple[int, int]] = None,
                 image_opts=None, cacheable: bool = True, georef=None):
        """
        :param image: the image
        :type image: PIL `Image`, image file object, or filename
        :param format: the format of the ``image``
        :param size: the size of the ``image`` in pixel
        """
        self._img: Optional[Image.Image] = None
        self._buf: Union[IO[bytes], ReadBufWrapper, None] = None
        self._fname: Optional[str] = None
        self.image = image
        self.image_opts = image_opts
        self._size = size
        self.cacheable = cacheable
        self.georef = georef
        self.authorize_stale = False

    @property
    def image(self):
        return self._img or self._buf or self._fname

    @image.setter
    def image(self, image):
        self._img = None
        self._buf = None
        if isinstance(image, str):
            self._fname = image
        elif isinstance(image, Image.Image):
            self._img = image
        else:
            self._buf = image

    def close_buffers(self):
        if self._buf:
            try:
                self._buf.close()
            except IOError:
                pass

    @property
    def filename(self):
        return self._fname

    def as_image(self) -> Image.Image:
        """
        Returns the image or the loaded image.

        :rtype: PIL `Image`
        """
        if not self._img:
            buf = self._seekable_buf
            log.debug('file(%s) -> image', self._fname or buf)

            try:
                img = Image.open(buf)
            except Exception:
                self.close_buffers()
                raise
            self._img = img
        if self.image_opts and self.image_opts.transparent and self._img.mode == 'P':
            self._img = self._img.convert('RGBA')
        return self._img

    @property
    def _seekable_buf(self) -> BytesIO:  # is this the correct type?
        if self._buf is None:
            if self._fname:
                self._buf = open(self._fname, 'rb')
            else:
                raise
        else:
            try:
                self._buf.seek(0)
            except (io.UnsupportedOperation, AttributeError):
                # PIL needs file objects with seek
                self._buf = BytesIO(self._buf.read())
        return cast(BytesIO, self._buf)

    @property
    def _readable_buf(self) -> Union[BytesIO, 'ReadBufWrapper']:  # is this the correct type?
        if self._buf is None:
            if self._fname:
                self._buf = open(self._fname, 'rb')
            else:
                raise
        elif not hasattr(self._buf, 'seek'):
            if not isinstance(self._buf, ReadBufWrapper):
                self._buf = ReadBufWrapper(self._buf)
        else:
            try:
                self._buf.seek(0)
            except (io.UnsupportedOperation, AttributeError):
                # PIL needs file objects with seek
                self._buf = BytesIO(self._buf.read())
        return cast(Union[BytesIO, 'ReadBufWrapper'], self._buf)

    def as_buffer(self, image_opts=None, format: Optional[str] = None, seekable: bool = False) ->\
            Union[IO[bytes], 'ReadBufWrapper']:
        """
        Returns the image as a file object.

        :param format: The format to encode an image.
                       Existing files will not be re-encoded.
        :rtype: file-like object
        """
        if format:
            image_opts = (image_opts or self.image_opts).copy()
            image_opts.format = ImageFormat(format)
        if not self._buf and not self._fname:
            if image_opts is None:
                image_opts = self.image_opts
            log.debug('image -> buf(%s)' % (image_opts.format,))
            assert self._img is not None
            self._buf = img_to_buf(self._img, image_opts=image_opts, georef=self.georef)
        else:
            buf = self._seekable_buf if seekable else self._readable_buf
            if self.image_opts and image_opts and not self.image_opts.format and image_opts.format:
                # need actual image_opts.format for next check
                self.image_opts = self.image_opts.copy()
                self.image_opts.format = peek_image_format(buf)
            if self.image_opts and image_opts and (
                    self.image_opts.format != image_opts.format or self.image_opts.mode != image_opts.mode):
                log.debug('converting image from %s -> %s' % (self.image_opts, image_opts))
                self.image = self.as_image()
                self._buf = None
                self.image_opts = image_opts
                # hide fname to prevent as_buffer from reading the file
                fname = self._fname
                self._fname = None
                self.as_buffer(image_opts)
                self._fname = fname
        assert self._buf is not None
        return self._buf

    @property
    def size(self):
        if self._size is None:
            self._size = self.as_image().size
        return self._size


def sub_image_result(image: BaseImageResult | Image.Image, size, offset, image_opts, cacheable=True) -> ImageResult:
    """
    Create a new ImageResult with `size` and `image_opts` and
    place `image` at `offset`.
    """
    # force new image to contain alpha channel
    new_image_opts = image_opts.copy()
    new_image_opts.transparent = True
    img = create_image(size, new_image_opts)

    if not hasattr(image, 'as_image'):
        image = ImageResult(image)
    subimg = image.as_image()
    img.paste(subimg, offset)
    return ImageResult(img, size=size, image_opts=new_image_opts, cacheable=cacheable)


class BlankImageResult(BaseImageResult):
    """
    ImageResult for transparent or solid-color images.
    Implements optimized as_buffer() method.
    """

    def __init__(self, size, image_opts, cacheable=False):
        self.size = size
        self.image_opts = image_opts
        self._buf = None
        self._img = None
        self.cacheable = cacheable
        self.authorize_stale = False

    def as_image(self) -> Image.Image:
        if not self._img:
            self._img = create_image(self.size, self.image_opts)
        return self._img

    def as_buffer(self, image_opts=None, format: Optional[str] = None, seekable: bool = False) -> \
            Union[IO[bytes], 'ReadBufWrapper']:
        if not self._buf:
            image_opts = (image_opts or self.image_opts).copy()
            if format:
                image_opts.format = ImageFormat(format)
            image_opts.colors = 0
            self._buf = img_to_buf(self.as_image(), image_opts=image_opts)
        return self._buf

    def close_buffers(self):
        pass


class ReadBufWrapper(object):
    """
    This class wraps everything with a ``read`` method and adds support
    for ``seek``, etc. A call to everything but ``read`` will create a
    StringIO object of the ``readbuf``.
    """

    def __init__(self, readbuf):
        self.ok_to_seek = False
        self.readbuf = readbuf
        self.stringio = None

    def read(self, *args, **kw):
        if self.stringio:
            return self.stringio.read(*args, **kw)
        return self.readbuf.read(*args, **kw)

    def __iter__(self):
        if self.stringio:
            return iter(self.stringio)
        else:
            return iter(self.readbuf)

    def __getattr__(self, name):
        if self.stringio is None:
            if hasattr(self.readbuf, name):
                return getattr(self.readbuf, name)
            elif name == '__length_hint__':
                raise AttributeError
            self.ok_to_seek = True
            self.stringio = BytesIO(self.readbuf.read())
        return getattr(self.stringio, name)


def img_has_transparency(img: Image.Image) -> bool:
    if img.mode == 'P':
        if img.info.get('transparency', False):
            return True
        # convert to RGBA and check alpha channel
        img = img.convert('RGBA')
    if img.mode == 'RGBA':
        # any alpha except fully opaque
        return any(img.histogram()[-256:-1])
    return False


def img_to_buf(img: Image.Image, image_opts, georef=None) -> BytesIO:
    defaults: dict[str, Any] = {}
    image_opts = image_opts.copy()

    # convert I or L images to target mode
    if image_opts.mode and img.mode[0] in ('I', 'L') and img.mode != image_opts.mode:
        img = img.convert(image_opts.mode)

    if (image_opts.colors is None and base_config().image.paletted
            and image_opts.format.endswith('png')):
        # force 255 colors for png with globals.image.paletted
        image_opts.colors = 255

    format = filter_format(image_opts.format.ext)
    if format == 'mixed':
        if img_has_transparency(img):
            format = 'png'
            image_opts.transparent = True
        else:
            format = 'jpeg'
            image_opts.colors = None
            image_opts.transparent = False

    # quantize if colors is set, but not if we already have a paletted image
    if image_opts.colors and not (img.mode == 'P' and len(img.getpalette() or []) == image_opts.colors*3):
        quantizer = None
        if 'quantizer' in image_opts.encoding_options:
            quantizer = image_opts.encoding_options['quantizer']
        if image_opts.transparent:
            img = quantize(img, colors=image_opts.colors, alpha=True,
                           defaults=defaults, quantizer=quantizer)
        else:
            img = quantize(img, colors=image_opts.colors,
                           quantizer=quantizer)
        if hasattr(Image, 'RLE'):
            defaults['compress_type'] = Image.RLE

    buf = BytesIO()
    if format == 'jpeg':
        img = img.convert('RGB')
        if 'jpeg_quality' in image_opts.encoding_options:
            defaults['quality'] = image_opts.encoding_options['jpeg_quality']
        else:
            defaults['quality'] = base_config().image.jpeg_quality

    elif format == 'tiff':
        if georef:
            tags = georef.tiff_tags(img.size)
            defaults['tiffinfo'] = tags
        if 'tiff_compression' in image_opts.encoding_options:
            defaults['compression'] = image_opts.encoding_options['tiff_compression']
            if defaults['compression'] == 'jpeg':
                if 'jpeg_quality' in image_opts.encoding_options:
                    defaults['quality'] = image_opts.encoding_options['jpeg_quality']

    # unsupported transparency tuple can still be in non-RGB img.infos
    # see: https://github.com/python-pillow/Pillow/pull/2633
    if (format == 'png' and img.mode != 'RGB'
            and 'transparency' in img.info and isinstance(img.info['transparency'], tuple)):
        del img.info['transparency']

    if image_opts.mode is not None and img.mode != image_opts.mode:
        img = img.convert(image_opts.mode)

    img.save(buf, format, **defaults)
    buf.seek(0)
    return buf


def quantize(img: Image.Image, colors=256, alpha=False, defaults=None, quantizer=None) -> Image.Image:
    if hasattr(Image, 'FASTOCTREE') and quantizer in (None, 'fastoctree'):
        if not alpha:
            img = img.convert('RGB')
        try:
            if img.mode == 'P':
                # quantize with alpha does not work with P images
                img = img.convert('RGBA')
            img = img.quantize(colors, Image.FASTOCTREE)
        except ValueError:
            pass
    else:
        palette = Image.ADAPTIVE  # type: ignore[attr-defined]
        if alpha and img.mode == 'RGBA':
            img.load()  # split might fail if image is not loaded
            alpha = img.split()[3]
            img = img.convert('RGB').convert('P', palette=palette, colors=colors-1)
            mask = Image.eval(alpha, lambda a: 255 if a <= 128 else 0)
            img.paste(255, mask)
            if defaults is not None:
                defaults['transparency'] = 255
        else:
            img = img.convert('RGB').convert('P', palette=palette, colors=colors)

    return img


def filter_format(format: str) -> str:
    if format.lower() == 'geotiff':
        return 'tiff'
    if format.lower().startswith('png'):
        return 'png'
    return format


# Pillow==8
image_filter = {
    'nearest': Image.NEAREST,  # type: ignore[attr-defined]
    'bilinear': Image.BILINEAR,  # type: ignore[attr-defined]
    'bicubic': Image.BICUBIC  # type: ignore[attr-defined]
}


def is_single_color_image(image: Image.Image) -> bool | tuple[int, ...] | float:
    """
    Checks if the `image` contains only one color.
    Returns ``False`` if it contains more than one color, else
    the color-tuple of the single color.
    """
    result = image.getcolors(1)
    # returns a list of (count, color), limit to one
    if result is None:
        return False

    color = result[0][1]
    if image.mode == 'P':
        assert isinstance(color, int)
        palette = image.getpalette()
        assert palette is not None
        return palette[color*3], palette[color*3+1], palette[color*3+2]

    return result[0][1]


def make_transparent(img: BaseImageResult, color, tolerance=10) -> ImageResult:
    """
    Create alpha channel for the given image and make each pixel
    in `color` full transparent.

    Returns an RGBA ImageSoruce.

    Modifies the image in-place, unless it needs to be converted
    first (P->RGB).

    :param color: RGB color tuple
    :param tolerance: tolerance applied to each color value
    """
    result = _make_transparent(img.as_image(), color, tolerance)
    image_opts = img.image_opts.copy()
    image_opts.transparent = True
    image_opts.mode = 'RGBA'
    return ImageResult(result, size=result.size, image_opts=image_opts)


def _make_transparent(img: Image.Image, color, tolerance=10) -> Image.Image:
    img.load()

    if img.mode == 'P':
        img = img.convert('RGBA')

    channels = img.split()
    mask_channels = []
    for ch, c in zip(channels, color):
        # create bit mask for each matched color
        low_c, high_c = c-tolerance, c+tolerance
        mask_channels.append(Image.eval(ch, lambda x: 255 if low_c <= x <= high_c else 0))

    # multiply channel bit masks to get a single mask
    alpha = reduce(ImageChops.multiply, mask_channels)
    # invert to get alpha channel
    alpha = ImageChops.invert(alpha)

    if len(channels) == 4:
        # multiply with existing alpha
        alpha = ImageChops.multiply(alpha, channels[-1])

    img.putalpha(alpha)
    return img


def bbox_position_in_image(bbox, size, src_bbox):
    """
    Calculate the position of ``bbox`` in an image of ``size`` and ``src_bbox``.
    Returns the sub-image size and the offset in pixel from top-left corner
    and the sub-bbox.

    >>> bbox_position_in_image((-180, -90, 180, 90), (600, 300), (-180, -90, 180, 90))
    ((600, 300), (0, 0), (-180, -90, 180, 90))
    >>> bbox_position_in_image((-200, -100, 200, 100), (600, 300), (-180, -90, 180, 90))
    ((540, 270), (30, 15), (-180, -90, 180, 90))
    >>> bbox_position_in_image((-200, -50, 200, 100), (600, 300), (-180, -90, 180, 90))
    ((540, 280), (30, 20), (-180, -50, 180, 90))
    >>> bbox_position_in_image((586400,196400,752800,362800), (256, 256), (586400,196400,752800,350000))
    ((256, 237), (0, 19), (586400, 196400, 752800, 350000))
    """
    coord_to_px = make_lin_transf(bbox, (0, 0) + size)
    offsets = [0, size[1], size[0], 0]
    sub_bbox = list(bbox)
    if src_bbox[0] > bbox[0]:
        sub_bbox[0] = src_bbox[0]
        x, y = coord_to_px((src_bbox[0], 0))
        offsets[0] = int(x)
    if src_bbox[1] > bbox[1]:
        sub_bbox[1] = src_bbox[1]
        x, y = coord_to_px((0, src_bbox[1]))
        offsets[1] = int(y)

    if src_bbox[2] < bbox[2]:
        sub_bbox[2] = src_bbox[2]
        x, y = coord_to_px((src_bbox[2], 0))
        offsets[2] = int(x)

    if src_bbox[3] < bbox[3]:
        sub_bbox[3] = src_bbox[3]
        x, y = coord_to_px((0, src_bbox[3]))
        offsets[3] = int(y)

    size = abs(offsets[2] - offsets[0]), abs(offsets[1] - offsets[3])
    return size, (offsets[0], offsets[3]), tuple(sub_bbox)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import re
import warnings

import yaml
ENV_PATTERN = re.compile(r"\$(\w+)|\$\{([^}]+)\}")

# Tracks variable names for which a warning has already been issued (once per name).
_warned_env_vars: set = set()


class YAMLError(Exception):
    pass


def load_yaml_file(file_or_filename):
    """
    Load yaml from file object or filename.
    """
    if isinstance(file_or_filename, str):
        with open(file_or_filename, 'rb') as f:
            return load_yaml(f)
    return load_yaml(file_or_filename)


def _load_yaml(doc):
    # try different methods to load yaml
    try:
        if getattr(yaml, '__with_libyaml__', False):
            try:
                return yaml.load(doc, Loader=yaml.CSafeLoader)
            except AttributeError:
                # handle cases where __with_libyaml__ is True but
                # CLoader doesn't work (missing .dispose())
                return yaml.safe_load(doc)
        return yaml.safe_load(doc)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as ex:
        raise YAMLError(str(ex))


def load_yaml(doc):
    """
    Load yaml from file object or string.
    """
    data = expand_env(_load_yaml(doc))
    if type(data) is not dict:
        raise YAMLError("configuration not a YAML dictionary")
    return data


# functions for using env-names in variables
def replace_env_vars(value):
    """Replaces $VAR and ${VAR} in a string.
    If the environment variable is not set, the original placeholder is kept
    as a fallback value and a :class:`UserWarning` is issued once per
    unknown variable name.
    """
    def repl(match):
        var_name = match.group(1) or match.group(2)
        if var_name in os.environ:
            return os.environ[var_name]
        fallback = match.group(0)
        if var_name not in _warned_env_vars:
            _warned_env_vars.add(var_name)
            warnings.warn(
                f"Environment variable '{var_name}' is not set. "
                f"Using fallback value: '{fallback}'",
                UserWarning,
                stacklevel=2,
            )
        return fallback

    return ENV_PATTERN.sub(repl, value)


def expand_env(obj):
    """Recursively traverse a nested Python object."""
    if isinstance(obj, dict):
        return {k: expand_env(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [expand_env(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(expand_env(v) for v in obj)
    elif isinstance(obj, str):
        return replace_env_vars(obj)
    else:
        return obj



from __future__ import division

from mapproxy.util.coverage import Coverage
from mapproxy.layer.map_layer import MapLayer


class LimitedLayer:
    """
    Wraps an existing layer temporary and stores additional
    attributes for geographical limits.
    """

    def __init__(self, layer: MapLayer, coverage: Coverage):
        self._layer = layer
        self.coverage = coverage

    def __getattr__(self, name):
        return getattr(self._layer, name)

    def combined_layer(self, other, query):
        if self.coverage == other.coverage:
            combined = self._layer.combined_layer(other, query)
            if combined:
                return LimitedLayer(combined, self.coverage)
        return None

    def get_info(self, query):
        if self.coverage:
            if not self.coverage.contains(query.coord, query.srs):
                return None
        return self._layer.get_info(query)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import operator
import threading
from abc import ABC, abstractmethod
from typing import Optional, Union

from shapely.geometry.base import BaseGeometry
from shapely.geometry import Point, MultiPolygon
from shapely.prepared import prep
from shapely.ops import unary_union

from mapproxy.util.bbox import bbox_intersects, bbox_contains
from mapproxy.util.py import cached_property
from mapproxy.util.geom import (
    load_polygon_lines,
    transform_geometry,
    bbox_polygon,
    EmptyGeometryError,
)
from mapproxy.srs import SRS, _SRS
from mapproxy.extent import MapExtent
from mapproxy.util.bbox import BBOX

import logging
from functools import reduce
log_config = logging.getLogger('mapproxy.config.coverage')


class Coverage(ABC):
    srs: _SRS
    bbox: BBOX
    clip = False

    @property
    @abstractmethod
    def extent(self) -> MapExtent:
        pass

    @property
    @abstractmethod
    def geom(self):
        pass

    @abstractmethod
    def transform_to(self, srs: _SRS) -> 'Coverage':
        pass

    @abstractmethod
    def intersects(self, bbox: BBOX, srs: _SRS) -> bool:
        pass

    @abstractmethod
    def contains(self, bbox: BBOX, srs: _SRS) -> bool:
        pass


class MultiCoverage(Coverage):
    """Aggregates multiple coverages"""

    def __init__(self, coverages: list[Coverage]):
        self.coverages = coverages
        self.bbox = self.extent.bbox

    @cached_property
    def extent(self) -> MapExtent:
        return reduce(operator.add, [c.extent for c in self.coverages])

    @property
    def geom(self):
        raise NotImplementedError('MultiCoverage does not have a geom')

    def intersects(self, bbox: BBOX, srs: _SRS):
        return any(c.intersects(bbox, srs) for c in self.coverages)

    def contains(self, bbox: BBOX, srs: _SRS):
        return any(c.contains(bbox, srs) for c in self.coverages)

    def transform_to(self, srs: _SRS) -> 'MultiCoverage':
        return MultiCoverage([c.transform_to(srs) for c in self.coverages])

    def __eq__(self, other):
        if not isinstance(other, MultiCoverage):
            return NotImplemented

        if self.bbox != other.bbox:
            return False

        if len(self.coverages) != len(other.coverages):
            return False

        for a, b in zip(self.coverages, other.coverages):
            if a != b:
                return False

        return True

    def __ne__(self, other):
        if not isinstance(other, MultiCoverage):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self):
        return '<MultiCoverage %r: %r>' % (self.extent.llbbox, self.coverages)


class BBOXCoverage(Coverage):
    def __init__(self, bbox, srs, clip=False):
        self.bbox = bbox
        self.srs = srs
        self.clip = clip

    @property
    def extent(self) -> MapExtent:
        return MapExtent(self.bbox, self.srs)

    @property
    def geom(self):
        return bbox_polygon(self.bbox)

    def _bbox_in_coverage_srs(self, bbox: BBOX, srs: _SRS) -> BBOX:
        if len(bbox) == 2:
            # point to bbox
            bbox = [bbox[0], bbox[1], bbox[0], bbox[1]]
        if srs != self.srs:
            bbox = srs.transform_bbox_to(self.srs, bbox)
        return bbox

    def intersects(self, bbox: BBOX, srs: _SRS) -> bool:
        bbox = self._bbox_in_coverage_srs(bbox, srs)
        return bbox_intersects(self.bbox, bbox)

    def intersection(self, bbox: BBOX, srs: _SRS) -> Optional['BBOXCoverage']:
        bbox = self._bbox_in_coverage_srs(bbox, srs)
        intersection = (
            max(self.bbox[0], bbox[0]),
            max(self.bbox[1], bbox[1]),
            min(self.bbox[2], bbox[2]),
            min(self.bbox[3], bbox[3]),
        )

        if intersection[0] >= intersection[2] or intersection[1] >= intersection[3]:
            return None
        return BBOXCoverage(intersection, self.srs, clip=self.clip)

    def contains(self, bbox: BBOX, srs: _SRS) -> bool:
        bbox = self._bbox_in_coverage_srs(bbox, srs)
        return bbox_contains(self.bbox, bbox)

    def transform_to(self, srs: _SRS) -> 'BBOXCoverage':
        if srs == self.srs:
            return self

        bbox = self.srs.transform_bbox_to(srs, self.bbox)
        return BBOXCoverage(bbox, srs, clip=self.clip)

    def __eq__(self, other):
        if not isinstance(other, BBOXCoverage):
            return NotImplemented

        if self.srs != other.srs:
            return False

        if self.bbox != other.bbox:
            return False

        return True

    def __ne__(self, other):
        if not isinstance(other, BBOXCoverage):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self):
        return '<BBOXCoverage %r/%r>' % (self.extent.llbbox, self.bbox)


class GeomCoverage(Coverage):
    def __init__(self, geom: BaseGeometry, srs, clip=False):
        self._geom = geom
        self.bbox = geom.bounds
        self.srs = srs
        self.clip = clip
        self._prep_lock = threading.Lock()
        self._prepared_geom = None
        self._prepared_counter = 0
        self._prepared_max = 10000

    @property
    def extent(self) -> MapExtent:
        return MapExtent(self.bbox, self.srs)

    @property
    def geom(self):
        return self._geom

    @property
    def prepared_geom(self):
        # GEOS internal data structure for prepared geometries grows over time,
        # recreate to limit memory consumption
        if not self._prepared_geom or self._prepared_counter > self._prepared_max:
            self._prepared_geom = prep(self.geom)
            self._prepared_counter = 0
        self._prepared_counter += 1
        return self._prepared_geom

    def _geom_in_coverage_srs(self, geom, srs):
        if isinstance(geom, BaseGeometry):
            if srs != self.srs:
                geom = transform_geometry(srs, self.srs, geom)
        elif len(geom) == 2:
            if srs != self.srs:
                geom = srs.transform_to(self.srs, geom)
            geom = Point(geom)
        else:
            if srs != self.srs:
                geom = srs.transform_bbox_to(self.srs, geom)
            geom = bbox_polygon(geom)
        return geom

    def transform_to(self, srs: _SRS) -> 'GeomCoverage':
        if srs == self.srs:
            return self

        geom = transform_geometry(self.srs, srs, self.geom)
        return GeomCoverage(geom, srs, clip=self.clip)

    def intersects(self, bbox: BBOX, srs: _SRS) -> bool:
        bbox = self._geom_in_coverage_srs(bbox, srs)
        with self._prep_lock:
            return self.prepared_geom.intersects(bbox)

    def intersection(self, bbox: BBOX, srs: _SRS) -> 'GeomCoverage':
        bbox = self._geom_in_coverage_srs(bbox, srs)
        return GeomCoverage(self.geom.intersection(bbox), self.srs, clip=self.clip)

    def contains(self, bbox: BBOX, srs: _SRS) -> bool:
        bbox = self._geom_in_coverage_srs(bbox, srs)
        with self._prep_lock:
            return self.prepared_geom.contains(bbox)

    def __eq__(self, other):
        if not isinstance(other, GeomCoverage):
            return NotImplemented

        if self.srs != other.srs:
            return False

        if self.bbox != other.bbox:
            return False

        if not self.geom.equals(other.geom):
            return False

        return True

    def __ne__(self, other):
        if not isinstance(other, GeomCoverage):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self):
        return '<GeomCoverage %r: %r>' % (self.extent.llbbox, self.geom)


def union_coverage(coverages: list[Coverage], clip=False) -> GeomCoverage:
    """
    Create a coverage that is the union of all `coverages`.
    Resulting coverage is in the SRS of the first coverage.
    """
    srs = coverages[0].srs

    coverages = [c.transform_to(srs) for c in coverages]

    geoms = [c.geom for c in coverages]

    union = unary_union(geoms)

    return GeomCoverage(union, srs=srs, clip=clip)


def coverage(geom: Union[BaseGeometry, list[float], tuple[float]], srs: _SRS, clip=False) -> 'Coverage':
    if isinstance(geom, (list, tuple)):
        return BBOXCoverage(bbox=geom, srs=srs, clip=clip)
    else:
        return GeomCoverage(geom=geom, srs=srs, clip=clip)


def load_limited_to(limited_to) -> 'GeomCoverage':
    srs = SRS(limited_to['srs'])
    geom = limited_to['geometry']

    if not hasattr(geom, 'geom_type'):  # not a Shapely geometry
        if isinstance(geom, (list, tuple)):
            geom = bbox_polygon(geom)
        else:
            polygons = load_polygon_lines(geom.split('\n'))
            if len(polygons) == 1:
                geom = polygons[0]
            else:
                geom = MultiPolygon(polygons)

    return GeomCoverage(geom, srs, clip=True)


def diff_coverage(coverages: list[Coverage], clip=False) -> GeomCoverage:
    """
    Create a coverage by subtracting all `coverages` from the first one.
    Resulting coverage is in the SRS of the first coverage.
    """
    srs = coverages[0].srs

    transformed_coverages = [c.transform_to(srs) for c in coverages]

    geoms = [c.geom for c in transformed_coverages]

    sub = unary_union(geoms[1:])
    diff = geoms[0].difference(sub)

    if diff.is_empty:
        raise EmptyGeometryError("diff did not return any geometry")

    return GeomCoverage(diff, srs=srs, clip=clip)


def intersection_coverage(coverages: list[GeomCoverage], clip=False) -> GeomCoverage:
    """
    Create a coverage by creating the intersection of all `coverages`.
    Resulting coverage is in the SRS of the first coverage.
    """
    srs = coverages[0].srs

    transformed_coverages = [c.transform_to(srs) for c in coverages]

    geoms = [c.geom for c in transformed_coverages]

    intersection = reduce(lambda a, b: a.intersection(b), geoms)

    if intersection.is_empty:
        raise EmptyGeometryError("intersection did not return any geometry")

    return GeomCoverage(intersection, srs=srs, clip=clip)



from contextlib import contextmanager
from functools import partial
from typing import Any, cast, Optional, Union

from mapproxy.cache.tile_creator import TileCreator
from mapproxy.image import BaseImageResult
from mapproxy.grid import TileCoord
from mapproxy.image import BlankImageResult
from mapproxy.cache.base import TileCacheBase
from mapproxy.cache.tile import Tile, TileCollection
from mapproxy.grid.meta_grid import MetaGrid
from mapproxy.grid.tile_grid import TileGrid
from mapproxy.image.mask import mask_image_result_from_coverage
from mapproxy.image.opts import ImageOptions
from mapproxy.image.tile import TiledImage
from mapproxy.layer.map_layer import MapLayer
from mapproxy.source import DummySource

# RESCALE_TILE_MISSING is a dummy image result to prevent a tile cache from loading
# a tile that we already found out is missing.

RESCALE_TILE_MISSING = BlankImageResult((256, 256), ImageOptions())


class TileManager:
    """
    Manages tiles for a single grid.
    Loads tiles from the cache, creates new tiles from sources and stores them
    into the cache, or removes tiles.

    :param pre_store_filter: a list with filter. each filter will be called
        with a tile before it will be stored to disc. the filter should
        return this or a new tile object.
    """

    def __init__(self, grid: TileGrid, cache: TileCacheBase, sources: list[MapLayer], format, locker, image_opts=None,
                 request_format=None, meta_buffer=None, meta_size=None, minimize_meta_requests=False, identifier=None,
                 pre_store_filter=None, concurrent_tile_creators=1, tile_creator_class=None,
                 bulk_meta_tiles=False, rescale_tiles=0, cache_rescaled_tiles=False, dimensions=None
                 ):
        self.grid = grid
        self.cache = cache
        self.locker = locker
        self.identifier = identifier
        self.meta_grid = None
        self.format = format
        self.image_opts = image_opts
        self.request_format = request_format or format
        self.sources = sources
        self.minimize_meta_requests = minimize_meta_requests
        self._expire_timestamp = None
        self._refresh_before: dict[str, Any] = {}
        self.pre_store_filter = pre_store_filter or []
        self.concurrent_tile_creators = concurrent_tile_creators
        self.tile_creator_class = tile_creator_class or TileCreator
        self.dimensions = dimensions

        self.rescale_tiles = rescale_tiles
        self.cache_rescaled_tiles = cache_rescaled_tiles

        if meta_buffer or (meta_size and not meta_size == [1, 1]):
            if all(source.supports_meta_tiles for source in sources):
                self.meta_grid = MetaGrid(grid, meta_size=meta_size, meta_buffer=meta_buffer)
            elif any(source.supports_meta_tiles for source in sources):
                raise ValueError('meta tiling configured but not supported by all sources')
            elif meta_size and not meta_size == [1, 1] and bulk_meta_tiles:
                # meta tiles configured but all sources are tiled
                # use bulk_meta_tile mode that download tiles in parallel
                self.meta_grid = MetaGrid(grid, meta_size=meta_size, meta_buffer=0)
                self.tile_creator_class = partial(self.tile_creator_class, bulk_meta_tiles=True)

    @contextmanager
    def session(self):
        """
        Context manager for access to the cache. Cleans up after usage
        for connection based caches.

        >>> with tile_manager.session(): #doctest: +SKIP
        ...    tile_manager.load_tile_coords(tile_coords)

        """
        yield
        self.cleanup()

    def cleanup(self):
        if hasattr(self.cache, 'cleanup'):
            self.cache.cleanup()

    def load_tile_coord(self, tile_coord: TileCoord, dimensions=None, with_metadata=False) -> Tile:
        return self.load_tile_coords(
            [tile_coord], dimensions=dimensions, with_metadata=with_metadata,
        )[0]

    def load_tile_coords(self, tile_coords: list[TileCoord], dimensions=None, with_metadata=False) -> TileCollection:
        tiles = TileCollection(tile_coords)
        rescale_till_zoom = 0
        if self.rescale_tiles:
            for t in tiles.tiles:
                # Use zoom level from first None tile.
                if t.coord is not None:
                    rescale_till_zoom = t.coord[2] + self.rescale_tiles
                    break
            else:
                return tiles

            if rescale_till_zoom < 0:
                rescale_till_zoom = 0
            if rescale_till_zoom > self.grid.levels:
                rescale_till_zoom = self.grid.levels

        # Remove tiles that are not in the cache coverage
        if self.cache.coverage:
            for t in tiles.tiles:
                if t.coord:
                    tile_bbox = self.grid.tile_bbox(t.coord)
                    if not self.cache.coverage.intersects(tile_bbox, self.grid.srs):
                        t.coord = None

        tiles = self._load_tile_coords(
            tiles, dimensions=dimensions, with_metadata=with_metadata,
            rescale_till_zoom=rescale_till_zoom, rescaled_tiles={},
        )

        for t in tiles.tiles:
            # Clip tiles if clipping is enabled for coverage
            if t.coord and self.cache.coverage and self.cache.coverage.clip and t.image_result:
                tile_bbox = self.grid.tile_bbox(t.coord)
                coverage = self.cache.coverage

                if coverage.intersects(tile_bbox, self.grid.srs):
                    t.image_result = mask_image_result_from_coverage(
                        t.image_result, tile_bbox, self.grid.srs, coverage,
                        ImageOptions(transparent=True, format='png'))

            # Remove our internal marker source, for missing tiles.
            if t.image_result is RESCALE_TILE_MISSING:
                t.image_result = None
        return tiles

    def _is_tile_missing(self, tile, cache_only, dimensions=None):
        if tile.coord is None:
            return False
        if cache_only:
            # in cache_only mode, we already fetched the tile from cache
            return tile.is_missing()
        else:
            # missing or staled
            return not self.is_cached(tile, dimensions=dimensions)

    def _load_tile_coords(self, tiles: TileCollection, dimensions=None, with_metadata=False,
                          rescale_till_zoom=None, rescaled_tiles=None
                          ) -> TileCollection:
        uncached_tiles = []

        if rescaled_tiles:
            for t in tiles:
                if t.coord in rescaled_tiles:
                    t.image_result = rescaled_tiles[t.coord].image_result

        # load all in batch
        self.cache.load_tiles(tiles, with_metadata, dimensions=dimensions)

        # if no real source, we are running in cache_only mode
        cache_only = self.sources == [] or (len(self.sources) == 1 and isinstance(self.sources[0], DummySource))
        # if no rescale_tiles and cache_only, we dont have any additional processing to do
        if self.rescale_tiles == 0 and cache_only:
            return tiles

        for tile in tiles:
            if self._is_tile_missing(tile, cache_only, dimensions=dimensions):
                uncached_tiles.append(tile)

        if uncached_tiles:
            creator = self.creator(dimensions=dimensions)
            created_tiles = creator.create_tiles(uncached_tiles)
            if not created_tiles and self.rescale_tiles:
                created_tiles = [self._scaled_tile(t, rescale_till_zoom, rescaled_tiles) for t in uncached_tiles]

            for created_tile in created_tiles:
                if created_tile.coord in tiles:
                    tiles[created_tile.coord].image_result = created_tile.image_result

        return tiles

    def remove_tile_coords(self, tile_coords: list[TileCoord]):
        tiles = TileCollection(tile_coords)
        self.cache.remove_tiles(tiles)

    def creator(self, dimensions=None):
        return self.tile_creator_class(self, dimensions=dimensions)

    def lock(self, tile):
        if self.meta_grid:
            tile = Tile(self.meta_grid.main_tile(tile.coord))
        return self.locker.lock(tile)

    def is_cached(self, tile: Union[Tile, TileCoord], dimensions=None) -> bool:
        """
        Return True if the tile is cached.
        """
        if isinstance(tile, tuple):
            tile = Tile(tile)
        if tile.coord is None:
            return True
        cached = self.cache.is_cached(tile, dimensions=dimensions)
        max_mtime = self.expire_timestamp()
        if cached and max_mtime is not None:
            self.cache.load_tile_metadata(tile, dimensions=self.dimensions)
            # file time stamp must be rounded to integer since time conversion functions
            # mktime and timetuple strip decimals from seconds
            assert tile.timestamp is not None
            stale = int(tile.timestamp) <= max_mtime
            if stale:
                cached = False
        return cached

    def is_stale(self, tile: Tile, dimensions=None) -> bool:
        """
        Return True if tile exists _and_ is expired.
        """
        if isinstance(tile, tuple):
            tile = Tile(cast(tuple[int, int, int], tile))
        if self.cache.is_cached(tile, dimensions=dimensions):
            # tile exists
            if not self.is_cached(tile, dimensions=dimensions):
                # expired
                return True
            return False
        return False

    def expire_timestamp(self):
        """
        Return the timestamp until which a tile should be accepted as up-to-date,
        or ``None`` if the tiles should not expire.

        :note: Returns _expire_timestamp by default.
        """
        if self._refresh_before:
            from mapproxy.seed.config import before_timestamp_from_options
            return before_timestamp_from_options(self._refresh_before)
        return self._expire_timestamp

    def apply_tile_filter(self, tile):
        """
        Apply all `pre_store_filter` to this tile.
        Returns filtered tile.
        """
        if tile.stored:
            return tile

        for img_filter in self.pre_store_filter:
            tile = img_filter(tile)
        return tile

    def _scaled_tile(self, tile: Tile, stop_zoom, rescaled_tiles) -> Tile:
        """
        Try to load tile by loading, scaling and clipping tiles from zoom levels above or
        below. stop_zoom determines if tiles from above should be scaled up, or if tiles
        from below should be scaled down.
        Returns an empty Tile if tile zoom level is stop_zoom.
        """
        if tile.coord in rescaled_tiles:
            return rescaled_tiles[tile.coord]

        # Cache tile in rescaled_tiles. We initially set source to a fixed
        # BlankImageResult and overwrite it if we actually rescaled the tile.
        tile.image_result = RESCALE_TILE_MISSING
        rescaled_tiles[tile.coord] = tile

        assert tile.coord is not None
        tile_bbox = self.grid.tile_bbox(tile.coord)
        current_zoom = tile.coord[2]
        if stop_zoom == current_zoom:
            return tile
        if stop_zoom > current_zoom:
            src_level = current_zoom + 1
        else:
            src_level = current_zoom - 1

        src_bbox, src_tile_grid, affected_tile_coords = self.grid.get_affected_level_tiles(tile_bbox, src_level)

        affected_tiles = TileCollection(affected_tile_coords)
        for t in affected_tiles:
            # Add sources of cached tiles, to avoid loading same tile multiple times
            # loading recursive.
            if t.coord in rescaled_tiles:
                t.image_result = rescaled_tiles[t.coord].image_result

        tile_collection = self._load_tile_coords(
            affected_tiles,
            rescale_till_zoom=stop_zoom,
            rescaled_tiles=rescaled_tiles,
        )

        if tile_collection.blank:
            return tile

        tile_results: list[Optional[BaseImageResult]] = []
        for t in tile_collection:
            # Replace RESCALE_TILE_MISSING with None, before transforming tiles.
            tile_results.append(t.image_result if t.image_result is not RESCALE_TILE_MISSING else None)

        tiled_image = TiledImage(tile_results, src_bbox=src_bbox, src_srs=self.grid.srs,
                                 tile_grid_size=src_tile_grid, tile_size=self.grid.tile_size)
        tile.image_result = tiled_image.transform(tile_bbox, self.grid.srs, self.grid.tile_size, self.image_opts)

        if self.cache_rescaled_tiles:
            self.cache.store_tile(tile)
        return tile



# This file is part of the MapProxy project.
# Copyright (C) 2010-2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import multiprocessing
import os
import random
import shutil
import sys
import tempfile
import threading
import time
from unittest.mock import patch

from mapproxy.util.lock import FileLock, SemLock, cleanup_lockdir, LockTimeout
from mapproxy.util.fs import (
    _force_rename_dir,
    swap_dir,
    cleanup_directory,
    write_atomic,
)
from mapproxy.util.py import reraise_exception
from mapproxy.util.times import timestamp_before
from mapproxy.test.helper import Mocker, assert_permissions

is_win = sys.platform == "win32"


class TestFileLock(Mocker):

    def setup_method(self):
        Mocker.setup_method(self)
        self.lock_dir = tempfile.mkdtemp()
        self.lock_file = os.path.join(self.lock_dir, "lock.lck")

    def teardown_method(self):
        shutil.rmtree(self.lock_dir)
        Mocker.teardown_method(self)

    def test_file_lock_timeout(self):
        lock = self._create_lock()
        assert_locked(self.lock_file)
        lock  # prevent lint warnings

    def test_file_lock(self):
        # Test a lock that becomes free during a waiting lock() call.
        class Lock(threading.Thread):

            def __init__(self, lock_file):
                super().__init__()
                self.lock_file = lock_file
                self.lock = FileLock(self.lock_file)

            def run(self):
                self.lock.lock()
                time.sleep(0.2)
                self.lock.unlock()

        lock_thread = Lock(self.lock_file)
        start_time = time.time()
        lock_thread.start()

        # wait until thread got the locked
        while not lock_thread.lock._locked:
            time.sleep(0.001)

        # one lock that times out
        assert_locked(self.lock_file)

        # one lock that will get it after some time
        x = FileLock(self.lock_file, timeout=0.3, step=0.001)
        x.lock()

        locked_for = time.time() - start_time
        assert locked_for - 0.2 <= 0.1, "locking took to long?! (rerun if not sure)"

        # cleanup
        x.unlock()
        lock_thread.join()

    def test_lock_cleanup(self):
        old_lock_file = os.path.join(self.lock_dir, "lock_old.lck")
        x = FileLock(old_lock_file)
        x.lock()
        x.unlock()
        mtime = os.stat(old_lock_file).st_mtime
        mtime -= 7 * 60
        os.utime(old_lock_file, (mtime, mtime))

        x = self._create_lock()
        x.unlock()
        assert os.path.exists(old_lock_file)
        assert os.path.exists(self.lock_file)
        cleanup_lockdir(self.lock_dir)

        assert not os.path.exists(old_lock_file)
        assert os.path.exists(self.lock_file)

    def test_concurrent_access(self):
        count_file = os.path.join(self.lock_dir, "count.txt")
        with open(count_file, "wb") as f:
            f.write(b"0")

        def count_up():
            with FileLock(self.lock_file, timeout=60):
                with open(count_file, "r+b") as f:
                    counter = int(f.read().strip())
                    f.seek(0)
                    f.write(str(counter + 1).encode("utf-8"))

        def do_it():
            for x in range(20):
                time.sleep(0.002)
                count_up()

        threads = [threading.Thread(target=do_it) for _ in range(20)]
        [t.start() for t in threads]
        [t.join() for t in threads]

        with open(count_file, "r+b") as f:
            counter = int(f.read().strip())

        assert counter == 400, counter

    def test_remove_on_unlock(self):
        x = FileLock(self.lock_file, remove_on_unlock=True)
        x.lock()
        assert os.path.exists(self.lock_file)
        x.unlock()
        if is_win:  # not removed on windows
            assert os.path.exists(self.lock_file)
        else:
            assert not os.path.exists(self.lock_file)

        if is_win:
            # not possible to remove lock file when lock is held
            pass
        else:
            x.lock()
            assert os.path.exists(self.lock_file)
            os.remove(self.lock_file)
            assert not os.path.exists(self.lock_file)
            # ignore removed lock
            x.unlock()
            assert not os.path.exists(self.lock_file)

    def test_file_lock_permissions(self):
        file_permissions = '775'

        # Test a lock that becomes free during a waiting lock() call.
        class Lock(threading.Thread):

            def __init__(self, lock_file):
                super().__init__()
                self.lock_file = lock_file
                self.lock = FileLock(self.lock_file, file_permissions=file_permissions)

            def run(self):
                self.lock.lock()
                time.sleep(0.2)
                self.lock.unlock()

        lock_thread = Lock(self.lock_file)
        lock_thread.start()

        # wait until thread got the locked
        while not lock_thread.lock._locked:
            time.sleep(0.001)

        # one lock that times out
        assert_locked(self.lock_file)
        assert_permissions(self.lock_file, file_permissions)

        lock_thread.join()

    def test_lock_retries_on_permission_error(self):
        real_open = open
        call_count = 0

        def flaky_open(path, *args, **kwargs):
            nonlocal call_count
            if path == self.lock_file:
                call_count += 1
                if call_count == 1:
                    raise PermissionError("mock transient failure")
            return real_open(path, *args, **kwargs)

        with patch("builtins.open", side_effect=flaky_open):
            lock = FileLock(self.lock_file, timeout=1.0, step=0.01)
            lock.lock()
            lock.unlock()

        assert call_count >= 2, "expected open to be called at least twice (initial fail + retry)"

    def test_lock_raises_lock_timeout_on_persistent_open_failure(self):
        real_open = open

        def always_fail(path, *args, **kwargs):
            if path == self.lock_file:
                raise PermissionError("mock persistent failure")
            return real_open(path, *args, **kwargs)

        with patch("builtins.open", side_effect=always_fail):
            lock = FileLock(self.lock_file, timeout=0.05, step=0.01)
            try:
                lock.lock()
            except LockTimeout:
                pass
            except PermissionError:
                assert False, "PermissionError should be wrapped as LockError and retried until LockTimeout"
            else:
                assert False, "expected LockTimeout"

    def _create_lock(self):
        lock = FileLock(self.lock_file)
        lock.lock()
        return lock


def assert_locked(lock_file, timeout=0.02, step=0.001):
    assert os.path.exists(lock_file)
    x = FileLock(lock_file, timeout=timeout, step=step)
    try:
        x.lock()
        assert False, "file was not locked"
    except LockTimeout:
        pass


class TestSemLock(object):

    def setup_method(self):
        self.lock_dir = tempfile.mkdtemp()
        self.lock_file = os.path.join(self.lock_dir, "lock.lck")

    def teardown_method(self):
        shutil.rmtree(self.lock_dir)

    def count_lockfiles(self):
        return len(glob.glob(self.lock_file + "*"))

    def test_single(self):
        locks = [SemLock(self.lock_file, 1, timeout=0.01) for _ in range(2)]
        locks[0].lock()
        try:
            locks[1].lock()
        except LockTimeout:
            pass
        else:
            assert False, "expected LockTimeout"

    def test_creating(self):
        locks = [SemLock(self.lock_file, 2) for _ in range(3)]

        assert self.count_lockfiles() == 0
        locks[0].lock()
        assert self.count_lockfiles() == 1
        locks[1].lock()
        assert self.count_lockfiles() == 2
        assert os.path.exists(locks[0]._lock._path)
        assert os.path.exists(locks[1]._lock._path)
        locks[0].unlock()
        locks[2].lock()
        locks[2].unlock()
        locks[1].unlock()

    def test_timeout(self):
        locks = [SemLock(self.lock_file, 2, timeout=0.1) for _ in range(3)]

        assert self.count_lockfiles() == 0
        locks[0].lock()
        assert self.count_lockfiles() == 1
        locks[1].lock()
        assert self.count_lockfiles() == 2
        try:
            locks[2].lock()
        except LockTimeout:
            pass
        else:
            assert False, "expected LockTimeout"
        locks[0].unlock()
        locks[2].unlock()

    def test_load(self):
        locks = [SemLock(self.lock_file, 8, timeout=1) for _ in range(20)]

        new_locks = random.sample([x for x in locks if not x._locked], 5)
        for x in new_locks:
            x.lock()

        for _ in range(20):
            old_locks = random.sample([x for x in locks if x._locked], 3)
            for x in old_locks:
                x.unlock()
            assert len([x for x in locks if x._locked]) == 2
            assert len([x for x in locks if not x._locked]) == 18

            new_locks = random.sample([x for x in locks if not x._locked], 3)
            for x in new_locks:
                x.lock()

            assert len([x for x in locks if x._locked]) == 5
            assert len([x for x in locks if not x._locked]) == 15

        assert self.count_lockfiles() == 8


class DirTest(object):

    def setup_method(self):
        self.tmpdir = tempfile.mkdtemp()

    def teardown_method(self):
        if os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def mkdir(self, name):
        dirname = os.path.join(self.tmpdir, name)
        os.mkdir(dirname)
        self.mkfile(name, dirname=dirname)
        return dirname

    def mkfile(self, name, dirname=None):
        if dirname is None:
            dirname = self.mkdir(name)
        filename = os.path.join(dirname, name + ".txt")
        open(filename, "wb").close()
        return filename


class TestForceRenameDir(DirTest):

    def test_rename(self):
        src_dir = self.mkdir("bar")
        dst_dir = os.path.join(self.tmpdir, "baz")
        _force_rename_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_rename_overwrite(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")
        _force_rename_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)


class TestSwapDir(DirTest):

    def test_swap_dir(self):
        src_dir = self.mkdir("bar")
        dst_dir = os.path.join(self.tmpdir, "baz")

        swap_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_swap_dir_w_old(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")

        swap_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_swap_dir_keep_old(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")

        swap_dir(src_dir, dst_dir, keep_old=True, backup_ext=".bak")
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert os.path.exists(dst_dir + ".bak")
        assert os.path.exists(os.path.join(dst_dir + ".bak", "baz.txt"))


class TestCleanupDirectory(DirTest):

    def test_no_remove(self):
        dirs = [self.mkdir("dir" + str(n)) for n in range(10)]
        for d in dirs:
            assert os.path.exists(d), d
        cleanup_directory(self.tmpdir, timestamp_before(minutes=1))
        for d in dirs:
            assert os.path.exists(d), d

    def test_file_handler(self):
        files = []
        file_handler_calls = []

        def file_handler(filename):
            file_handler_calls.append(filename)

        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            os.utime(filename, (new_date, new_date))
            files.append(filename)

        for filename in files:
            assert os.path.exists(filename), filename
        cleanup_directory(self.tmpdir, timestamp_before(), file_handler=file_handler)
        for filename in files:
            assert os.path.exists(filename), filename

        assert set(files) == set(file_handler_calls)

    def test_no_directory(self):
        cleanup_directory(os.path.join(self.tmpdir, "invalid"), timestamp_before())
        # nothing should happen

    def test_remove_all(self):
        files = []
        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            os.utime(filename, (new_date, new_date))
            files.append(filename)

        for filename in files:
            assert os.path.exists(filename), filename
        cleanup_directory(self.tmpdir, timestamp_before())
        for filename in files:
            assert not os.path.exists(filename), filename
            assert not os.path.exists(os.path.dirname(filename)), filename

    def test_remove_empty_dirs(self):
        os.makedirs(os.path.join(self.tmpdir, "foo", "bar", "baz"))
        cleanup_directory(self.tmpdir, timestamp_before(minutes=-1))
        assert not os.path.exists(os.path.join(self.tmpdir, "foo"))

    def test_remove_some(self):
        files = []
        # create a few files, every other file is one week old
        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            if n % 2 == 0:
                os.utime(filename, (new_date, new_date))
            files.append(filename)

        # check all files are present
        for filename in files:
            assert os.path.exists(filename), filename

        # cleanup_directory for all files older then one minute
        cleanup_directory(self.tmpdir, timestamp_before(minutes=1))

        # check old files and dirs are removed
        for filename in files[::2]:
            assert not os.path.exists(filename), filename
            assert not os.path.exists(os.path.dirname(filename)), filename

        # check new files are still present
        for filename in files[1::2]:
            assert os.path.exists(filename), filename


def _write_atomic_data(i_filename):
    (i, filename) = i_filename
    data = str(i) + "\n" + "x" * 10000
    write_atomic(filename, data.encode("utf-8"))
    time.sleep(0.001)


class TestWriteAtomic(object):

    def setup_method(self):
        self.dirname = tempfile.mkdtemp()

    def teardown_method(self):
        if self.dirname:
            shutil.rmtree(self.dirname)

    def test_concurrent_write(self):
        filename = os.path.join(self.dirname, "tmpfile")

        num_writes = 800
        concurrent_writes = 8

        p = multiprocessing.Pool(concurrent_writes)
        p.map(_write_atomic_data, ((i, filename) for i in range(num_writes)))
        p.close()
        p.join()

        assert os.path.exists(filename)
        last_i = int(open(filename).readline())
        assert last_i > (num_writes / 2), (
            "file should contain content from "
            "later writes, got content from write %d" % (last_i + 1)
        )
        os.unlink(filename)
        assert os.listdir(self.dirname) == []

    def test_not_a_file(self):
        # check that expected errors are not hidden
        filename = os.path.join(self.dirname, "tmpfile")
        os.mkdir(filename)

        try:
            write_atomic(filename, b"12345")
        except (OSError, IOError):
            pass
        else:
            assert False, "expected exception"


def test_reraise_exception():

    def valueerror_raiser():
        raise ValueError()

    def reraiser():
        try:
            valueerror_raiser()
        except ValueError:
            raise reraise_exception(TypeError(), sys.exc_info())

    try:
        reraiser()
    except TypeError as ex:
        assert ex
    else:
        assert False, "expected exception"



from __future__ import division

from typing import Optional

from mapproxy.cache.tile_manager import TileManager
from mapproxy.extent import map_extent_from_grid, MapExtent
from mapproxy.grid import NoTiles, GridError
from mapproxy.image import bbox_position_in_image, sub_image_result, BaseImageResult
from mapproxy.image.tile import TiledImage
from mapproxy.layer import merge_layer_res_ranges, BlankImageError, MapError, MapBBOXError
from mapproxy.layer.map_layer import MapLayer
from mapproxy.proj import ProjError
from mapproxy.query import MapQuery
from mapproxy.util.bbox import bbox_equals


class CacheMapLayer(MapLayer):
    supports_meta_tiles = True

    def __init__(self, tile_manager: TileManager, extent: Optional[MapExtent] = None, image_opts=None,
                 max_tile_limit=None):
        super().__init__(image_opts=image_opts)
        self.tile_manager = tile_manager
        self.grid = tile_manager.grid
        self.extent = extent or map_extent_from_grid(self.grid)
        self.res_range = None
        if not self.tile_manager.rescale_tiles:
            self.res_range = merge_layer_res_ranges(self.tile_manager.sources)
        self.max_tile_limit = max_tile_limit

    def get_map(self, query: MapQuery) -> BaseImageResult:
        self.check_res_range(query)

        if query.tiled_only:
            self._check_tiled(query)

        query_extent = MapExtent(query.bbox, query.srs)
        if not query.tiled_only and self.extent and not self.extent.contains(query_extent):
            if not self.extent.intersects(query_extent):
                raise BlankImageError()
            size, offset, bbox = bbox_position_in_image(query.bbox, query.size, self.extent.bbox_for(query.srs))
            if size[0] == 0 or size[1] == 0:
                raise BlankImageError()
            src_query = MapQuery(bbox, size, query.srs, query.format, dimensions=query.dimensions)
            resp = self._image(src_query)
            return sub_image_result(resp, size=query.size, offset=offset, image_opts=self.image_opts,
                                    cacheable=resp.cacheable)
        else:
            return self._image(query)

    def _check_tiled(self, query):
        if query.format != self.tile_manager.format:
            raise MapError("invalid tile format, use %s" % self.tile_manager.format)
        if query.size != self.grid.tile_size:
            raise MapError("invalid tile size (use %dx%d)" % self.grid.tile_size)

    def _image(self, query: MapQuery) -> BaseImageResult:
        try:
            src_bbox, tile_grid_size, affected_tile_coords = \
                self.grid.get_affected_tiles(query.bbox, query.size,
                                             req_srs=query.srs)
        except NoTiles:
            raise BlankImageError()
        except GridError as ex:
            raise MapBBOXError(ex.args[0])

        num_tiles = tile_grid_size[0] * tile_grid_size[1]

        if self.max_tile_limit and num_tiles >= self.max_tile_limit:
            raise MapBBOXError("too many tiles, max_tile_limit: %s, num_tiles: %s" % (self.max_tile_limit, num_tiles))

        if query.tiled_only:
            if num_tiles > 1:
                raise MapBBOXError("not a single tile")
            bbox = query.bbox
            if not bbox_equals(bbox, src_bbox, abs((bbox[2]-bbox[0])/query.size[0]/10),
                               abs((bbox[3]-bbox[1])/query.size[1]/10)):
                raise MapBBOXError("query does not align to tile boundaries")

        with self.tile_manager.session():
            tile_collection = self.tile_manager.load_tile_coords(
                affected_tile_coords, with_metadata=query.tiled_only, dimensions=query.dimensions)

        if tile_collection.empty:
            raise BlankImageError()

        if query.tiled_only:
            tile = tile_collection[0].image_result
            tile.image_opts = self.tile_manager.image_opts
            tile.cacheable = tile_collection[0].cacheable
            return tile

        tile_image_results = [tile.image_result for tile in tile_collection]
        tiled_image = TiledImage(tile_image_results, src_bbox=src_bbox, src_srs=self.grid.srs,
                                 tile_grid_size=tile_grid_size, tile_size=self.grid.tile_size)
        try:
            return tiled_image.transform(query.bbox, query.srs, query.size,
                                         self.tile_manager.image_opts)
        except ProjError:
            raise MapBBOXError("could not transform query BBOX")
        except IOError as ex:
            from mapproxy.source import SourceError
            raise SourceError("unable to transform image: %s" % ex)



from __future__ import division

import os
import warnings
from collections import OrderedDict
from copy import deepcopy

from mapproxy.config import defaults
from mapproxy.config.configuration.service import ServiceConfiguration
from mapproxy.config.configuration.layer import WMSLayerConfiguration, LayerConfiguration
from mapproxy.config.configuration.cache import CacheConfiguration
from mapproxy.config.configuration.source import SourcesCollection, SourceConfiguration
from mapproxy.config.configuration.global_conf import GlobalConfiguration
from mapproxy.config.configuration.grid import GridConfiguration


class ProxyConfiguration(object):
    wms_root_layer: WMSLayerConfiguration

    def __init__(self, conf, conf_base_dir=None, seed=False, renderd=False):
        self.configuration = conf
        self.seed = seed
        self.renderd = renderd

        if conf_base_dir is None:
            conf_base_dir = os.getcwd()

        self.load_globals(conf_base_dir=conf_base_dir)
        self.load_grids()
        self.load_caches()
        self.load_sources()
        self.load_wms_root_layer()
        self.load_tile_layers()
        self.load_services()

    def load_globals(self, conf_base_dir):
        self.globals = GlobalConfiguration(conf_base_dir=conf_base_dir,
                                           conf=self.configuration.get('globals') or {},
                                           context=self)

    def load_grids(self):
        self.grids = {}
        grid_configs = dict(defaults.grids)
        grid_configs.update(self.configuration.get('grids') or {})
        for grid_name, grid_conf in grid_configs.items():
            grid_conf.setdefault('name', grid_name)
            self.grids[grid_name] = GridConfiguration(grid_conf, context=self)

    def load_caches(self):
        self.caches = OrderedDict()
        caches_conf = self.configuration.get('caches')
        if not caches_conf:
            return
        if isinstance(caches_conf, list):
            caches_conf = list_of_dicts_to_ordered_dict(caches_conf)
        for cache_name, cache_conf in caches_conf.items():
            cache_conf['name'] = cache_name
            self.caches[cache_name] = CacheConfiguration(conf=cache_conf, context=self)

    def load_sources(self):
        self.sources = SourcesCollection()
        for source_name, source_conf in (self.configuration.get('sources') or {}).items():
            source_conf['name'] = source_name
            self.sources[source_name] = SourceConfiguration.load(conf=source_conf, context=self)

    def load_tile_layers(self):
        self.layers = OrderedDict()
        layers_conf = deepcopy(self._layers_conf_dict())
        if layers_conf is None:
            return
        layers = self._flatten_layers_conf_dict(layers_conf)
        for layer_name, layer_conf in layers.items():
            layer_conf['name'] = layer_name
            self.layers[layer_name] = LayerConfiguration(conf=layer_conf, context=self)

    def _legacy_layers_conf_dict(self):
        """
        Read old style layer configuration with a dictionary where
        the key is the layer name. Optionally: a list an each layer
        is wrapped in such dictionary.

        ::
          layers:
            foo:
              title: xxx
              sources: []
            bar:
              title: xxx
              sources: []

        or

        ::

          layers:
            - foo:
               title: xxx
               sources: []
            - bar:
               title: xxx
               sources: []

        """
        warnings.warn('old layer configuration syntax is deprecated since 1.4.0. '
                      'use list of dictionaries as documented', RuntimeWarning)
        layers = []
        layers_conf = self.configuration.get('layers')
        if not layers_conf:
            return None  # TODO config error
        if isinstance(layers_conf, list):
            layers_conf = list_of_dicts_to_ordered_dict(layers_conf)
        for layer_name, layer_conf in layers_conf.items():
            layer_conf['name'] = layer_name
            layers.append(layer_conf)
        return dict(title=None, layers=layers)

    def _layers_conf_dict(self):
        """
        Returns (recursive) layer configuration as a dictionary
        in unified structure:

        ::
            {
             title: 'xxx', # required, might be None
             name: 'xxx', # optional
             # sources or layers or both are required
             sources: [],
             layers: [
                {..., ...} # more layers like this
             ]
            }

        Multiple layers will be wrapped in an unnamed root layer, if the
        first level starts with multiple layers.
        """
        layers_conf = self.configuration.get('layers')
        if layers_conf is None:
            return

        if isinstance(layers_conf, list):
            if isinstance(layers_conf[0], dict) and len(layers_conf[0].keys()) == 1:
                # looks like ordered legacy config
                layers_conf = self._legacy_layers_conf_dict()
            elif len(layers_conf) == 1 and (
                'layers' in layers_conf[0]
                or 'sources' in layers_conf[0]
                    or 'tile_sources' in layers_conf[0]):
                # single root layer in list -> remove list
                layers_conf = layers_conf[0]
            else:
                # layer list without root -> wrap in root layer
                layers_conf = dict(title=None, layers=layers_conf)

        if len(layers_conf.keys() &
               set('layers name title sources'.split())) < 2:
            # looks like unordered legacy config
            layers_conf = self._legacy_layers_conf_dict()

        return layers_conf

    def _flatten_layers_conf_dict(self, layers_conf, _layers=None):
        """
        Returns a dictionary with all layers that have a name and sources.
        Flattens the layer tree.
        """
        layers = _layers if _layers is not None else OrderedDict()

        if 'layers' in layers_conf:
            for layer in layers_conf.pop('layers'):
                self._flatten_layers_conf_dict(layer, layers)

        if 'name' in layers_conf and ('sources' in layers_conf or 'tile_sources' in layers_conf):
            layers[layers_conf['name']] = layers_conf

        return layers

    def load_wms_root_layer(self):
        self.wms_root_layer = None

        layers_conf = self._layers_conf_dict()
        if layers_conf is None:
            return
        self.wms_root_layer = WMSLayerConfiguration(layers_conf, context=self)

    def load_services(self):
        self.services = ServiceConfiguration(self.configuration.get('services', {}), context=self)

    def configured_services(self):
        with self:
            return self.services.services()

    def __enter__(self):
        # push local base_config onto config stack
        import mapproxy.config.config
        mapproxy.config.config._config.push(self.base_config)

    def __exit__(self, type, value, traceback):
        # pop local base_config from config stack
        import mapproxy.config.config
        mapproxy.config.config._config.pop()

    @property
    def base_config(self):
        return self.globals.base_config

    def config_files(self):
        """
        Returns a dictionary with all configuration filenames and there timestamps.
        Contains any included files as well (see `base` option).
        """
        return self.configuration.get('__config_files__', {})


def list_of_dicts_to_ordered_dict(dictlist):
    """
    >>> d = list_of_dicts_to_ordered_dict([{'a': 1}, {'b': 2}, {'c': 3}])
    >>> list(d.items())
    [('a', 1), ('b', 2), ('c', 3)]
    """

    result = OrderedDict()
    for d in dictlist:
        for k, v in d.items():
            result[k] = v
    return result



from __future__ import division

from mapproxy.util.bbox import BBOX
from mapproxy.srs import make_lin_transf, _SRS


class MapQuery:
    """
    Internal query for a map with a specific extent, size, srs, etc.
    """

    def __init__(self, bbox: BBOX, size: tuple[int, int], srs: _SRS, format='image/png', transparent=False,
                 tiled_only=False, dimensions=None):
        self.bbox = bbox
        self.size = size
        self.srs = srs
        self.format = format
        self.transparent = transparent
        self.tiled_only = tiled_only
        self.dimensions = dimensions or {}

    def dimensions_for_params(self, params):
        """
        Return subset of the dimensions.

        >>> mq = MapQuery(None, None, None, dimensions={'Foo': 1, 'bar': 2})
        >>> mq.dimensions_for_params(set(['FOO', 'baz']))
        {'Foo': 1}
        """
        params = [p.lower() for p in params]
        return dict((k, v) for k, v in self.dimensions.items() if k.lower() in params)

    def __repr__(self):
        info = self.__dict__
        serialized_dimensions = ", ".join(["'%s': '%s'" % (key, value) for (key, value) in self.dimensions.items()])
        info["serialized_dimensions"] = serialized_dimensions
        return ("MapQuery(bbox=%(bbox)s, size=%(size)s, srs=%(srs)r, format=%(format)s,"
                " dimensions={%(serialized_dimensions)s)}") % info


class InfoQuery(object):
    def __init__(self, bbox, size, srs, pos, info_format, format=None,
                 feature_count=None):
        self.bbox = bbox
        self.size = size
        self.srs = srs
        self.pos = pos
        self.info_format = info_format
        self.format = format
        self.feature_count = feature_count

    @property
    def coord(self):
        return make_lin_transf((0, 0, self.size[0], self.size[1]), self.bbox)(self.pos)


class LegendQuery(object):
    def __init__(self, format, scale):
        self.format = format
        self.scale = scale



from __future__ import division

import os

from mapproxy.config import load_default_config, finish_base_config
from mapproxy.config.configuration.base import ConfigurationBase
from mapproxy.config.configuration.base import dotted_dict_get
from mapproxy.config.configuration.image_options import ImageOptionsConfiguration


def preferred_srs(conf):
    from mapproxy.srs import SRS, PreferredSrcSRS

    preferred_conf = conf.get('preferred_src_proj', {})

    if not preferred_conf:
        return

    preferred = PreferredSrcSRS()

    for target, preferred_srcs in preferred_conf.items():
        preferred.add(SRS(target), [SRS(s) for s in preferred_srcs])

    return preferred


class GlobalConfiguration(ConfigurationBase):
    def __init__(self, conf_base_dir, conf, context):
        super().__init__(conf, context)
        self.base_config = load_default_config()
        self._copy_conf_values(self.conf, self.base_config)
        self.base_config.conf_base_dir = conf_base_dir
        finish_base_config(self.base_config)

        self.image_options = ImageOptionsConfiguration(self.conf.get('image', {}), context)
        self.preferred_srs = preferred_srs(self.conf.get('srs', {}))
        self.renderd_address = self.get_value('renderd.address')

    def _copy_conf_values(self, d, target):
        for k, v in d.items():
            if v is None:
                continue
            if hasattr(v, 'items') and k in target:
                self._copy_conf_values(v, target[k])
            else:
                target[k] = v

    def get_value(self, key, local={}, global_key=None, default_key=None):
        result = dotted_dict_get(key, local)
        if result is None:
            result = dotted_dict_get(global_key or key, self.conf)

        if result is None:
            result = dotted_dict_get(default_key or global_key or key, self.base_config)

        return result

    def get_path(self, key, local, global_key=None, default_key=None):
        value = self.get_value(key, local, global_key, default_key)
        if value is not None:
            value = self.abspath(value)
        return value

    def abspath(self, path):
        return os.path.join(self.base_config.conf_base_dir, path)



import sys
from typing import Optional, Callable, TypeVar, Union, TYPE_CHECKING

from mapproxy.cache.tile import TileCollection
from mapproxy.grid import TileCoord
from mapproxy.grid.meta_grid import MetaTile
from mapproxy.cache.tile import Tile
from mapproxy.image import BaseImageResult
from mapproxy.image.merge import merge_images
from mapproxy.image.tile import TileSplitter
from mapproxy.layer import BlankImageError
from mapproxy.layer.map_layer import MapLayer
from mapproxy.query import MapQuery
from mapproxy.source import SourceError
from mapproxy.util import async_
from mapproxy.util.coverage import Coverage
from mapproxy.util.py import reraise_exception, reraise
if TYPE_CHECKING:
    from mapproxy.cache.tile_manager import TileManager

# TypeVar for _create_threaded method
TTile = TypeVar('TTile', bound=Union[Tile, MetaTile])


class TileCreator:
    def __init__(self, tile_mgr: 'TileManager', dimensions=None, image_merger=None, bulk_meta_tiles=False):
        self.cache = tile_mgr.cache
        self.sources = tile_mgr.sources
        self.grid = tile_mgr.grid
        self.meta_grid = tile_mgr.meta_grid
        self.bulk_meta_tiles = bulk_meta_tiles
        self.tile_mgr = tile_mgr
        self.dimensions = dimensions
        self.image_merger = image_merger

    def is_cached(self, tile: Union[Tile, TileCoord], dimensions=None) -> bool:
        """
        Return True if the tile is cached.
        """
        return self.tile_mgr.is_cached(tile, dimensions=dimensions)

    def is_stale(self, tile: Tile) -> bool:
        """
        Return True if the tile exists in cache and is expired.
        """
        return self.tile_mgr.is_stale(tile)

    def create_tiles(self, tiles: list[Tile]) -> list[Tile]:
        if not self.sources:
            return []
        if not self.meta_grid:
            created_tiles = self._create_single_tiles(tiles)
        elif self.tile_mgr.minimize_meta_requests and len(tiles) > 1:
            # use minimal requests only for multiple tile requests (ie not for TMS)
            meta_tile = self.meta_grid.minimal_meta_tile([t.coord for t in tiles])
            created_tiles = self._create_meta_tile(meta_tile)
        else:
            meta_tiles = []
            meta_bboxes = set()
            for tile in tiles:
                assert tile.coord is not None
                meta_tile = self.meta_grid.meta_tile(tile.coord)
                if meta_tile.bbox not in meta_bboxes:
                    meta_tiles.append(meta_tile)
                    meta_bboxes.add(meta_tile.bbox)

            created_tiles = self._create_meta_tiles(meta_tiles)

        return created_tiles

    def _create_single_tiles(self, tiles: list[Tile]) -> list[Tile]:
        if self.tile_mgr.concurrent_tile_creators > 1 and len(tiles) > 1:
            return self._create_threaded(self._create_single_tile, tiles)

        created_tiles = []
        for tile in tiles:
            created_tiles.extend(self._create_single_tile(tile))
        return created_tiles

    def _create_threaded(self, create_func: Callable[[TTile], list[Tile]], tiles: list[TTile]) -> list[Tile]:
        result = []
        async_pool = async_.Pool(self.tile_mgr.concurrent_tile_creators)
        for new_tiles in async_pool.imap(create_func, tiles):
            result.extend(new_tiles)
        return result

    def _create_single_tile(self, tile: Tile, dimensions=None) -> list[Tile]:
        assert tile.coord is not None
        tile_bbox = self.grid.tile_bbox(tile.coord)
        query = MapQuery(tile_bbox, self.grid.tile_size, self.grid.srs,
                         self.tile_mgr.request_format, dimensions=self.dimensions)
        with self.tile_mgr.lock(tile):
            if not self.is_cached(tile, dimensions=dimensions):
                image_result = None
                try:
                    image_result = self._query_sources(query)
                # if source is not available, try to serve tile in cache
                except SourceError as e:
                    if self.is_stale(tile):
                        self.cache.load_tile(tile)
                    else:
                        raise reraise_exception(e, sys.exc_info())
                if not image_result:
                    return []
                if image_result.authorize_stale and self.is_stale(tile):
                    # The configuration authorises blank tiles generated by the error_handler
                    # to be replaced by stale tiles from cache.
                    self.cache.load_tile(tile)
                    return [tile]
                if self.tile_mgr.image_opts != image_result.image_opts:
                    # call as_buffer to force conversion into cache format
                    image_result.as_buffer(self.tile_mgr.image_opts)
                image_result.image_opts = self.tile_mgr.image_opts
                tile.image_result = image_result
                tile.cacheable = image_result.cacheable
                tile = self.tile_mgr.apply_tile_filter(tile)
                if image_result.cacheable:
                    self.cache.store_tile(tile)
            else:
                self.cache.load_tile(tile)
        return [tile]

    def _query_sources(self, query: MapQuery) -> Optional[BaseImageResult]:
        """
        Query all sources and return the results as a single ImageResult.
        Multiple sources will be merged into a single image.
        """

        # directly return get_map without merge if ...
        if (len(self.sources) == 1 and
                not self.image_merger and  # no special image_merger (like BandMerger)
                not (self.sources[0].coverage and  # no clipping coverage
                     self.sources[0].coverage.clip and
                     self.sources[0].coverage.intersects(query.bbox, query.srs))):
            try:
                return self.sources[0].get_map(query)
            except BlankImageError:
                return None

        def get_map_from_source(source: MapLayer) -> tuple[Optional[BaseImageResult], Optional[Coverage]]:
            try:
                img = source.get_map(query)
            except BlankImageError:
                return None, None
            else:
                return img, source.coverage

        layers: list[tuple[BaseImageResult, Coverage]] = []
        for layer in async_.imap(get_map_from_source, self.sources):
            if layer[0] is not None:
                layers.append(layer)

        return merge_images(layers, size=query.size, bbox=query.bbox, bbox_srs=query.srs,
                            image_opts=self.tile_mgr.image_opts, merger=self.image_merger)

    def _create_meta_tiles(self, meta_tiles: list[MetaTile]) -> list[Tile]:
        if self.bulk_meta_tiles:
            created_tiles = []
            for meta_tile in meta_tiles:
                created_tiles.extend(self._create_bulk_meta_tile(meta_tile))
            return created_tiles

        if self.tile_mgr.concurrent_tile_creators > 1 and len(meta_tiles) > 1:
            return self._create_threaded(self._create_meta_tile, meta_tiles)

        created_tiles = []
        for meta_tile in meta_tiles:
            created_tiles.extend(self._create_meta_tile(meta_tile))
        return created_tiles

    def _create_meta_tile(self, meta_tile: MetaTile) -> list[Tile]:
        """
        _create_meta_tile queries a single meta tile and splits it into
        tiles.
        """
        tile_size = self.grid.tile_size
        query = MapQuery(meta_tile.bbox, meta_tile.size, self.grid.srs, self.tile_mgr.request_format,
                         dimensions=self.dimensions)
        main_tile = Tile(meta_tile.main_tile_coord)
        with self.tile_mgr.lock(main_tile):
            if not all(self.is_cached(t, dimensions=self.dimensions) for t in meta_tile.tiles if t is not None):
                meta_tile_image = self._query_sources(query)
                if not meta_tile_image:
                    return []
                splitted_tiles = split_meta_tiles(meta_tile_image, meta_tile.tile_patterns,
                                                  tile_size, self.tile_mgr.image_opts)
                splitted_tiles = [self.tile_mgr.apply_tile_filter(t) for t in splitted_tiles]
                if meta_tile_image.cacheable:
                    self.cache.store_tiles(splitted_tiles, dimensions=self.dimensions)
                return splitted_tiles
            # else
        tiles = TileCollection(meta_tile.tiles)
        self.cache.load_tiles(tiles, dimensions=self.dimensions)
        return tiles.tiles

    def _create_bulk_meta_tile(self, meta_tile):
        """
        _create_bulk_meta_tile queries each tile of the meta tile in parallel
        (using concurrent_tile_creators).
        """
        tile_size = self.grid.tile_size
        main_tile = Tile(meta_tile.main_tile_coord)
        with self.tile_mgr.lock(main_tile):
            if not all(self.is_cached(t, dimensions=self.dimensions) for t in meta_tile.tiles if t is not None):
                async_pool = async_.Pool(self.tile_mgr.concurrent_tile_creators)

                def query_tile(coord):
                    try:
                        query = MapQuery(
                            self.grid.tile_bbox(coord), tile_size, self.grid.srs, self.tile_mgr.request_format,
                            dimensions=self.dimensions)
                        tile_image = self._query_sources(query)
                        if tile_image is None:
                            return None

                        if self.tile_mgr.image_opts != tile_image.image_opts:
                            # call as_buffer to force conversion into cache format
                            tile_image.as_buffer(self.tile_mgr.image_opts)

                        tile = Tile(coord, cacheable=tile_image.cacheable)
                        tile.image_result = tile_image
                        tile = self.tile_mgr.apply_tile_filter(tile)
                    except BlankImageError:
                        return None
                    else:
                        return tile

                tiles = []
                for tile_task in async_pool.imap(query_tile,
                                                 [t for t in meta_tile.tiles if t is not None],
                                                 use_result_objects=True,
                                                 ):
                    if tile_task.exception is None:
                        tile = tile_task.result
                        if tile is not None:
                            tiles.append(tile)
                    else:
                        ex = tile_task.exception
                        async_pool.shutdown(True)
                        raise reraise(ex)

                self.cache.store_tiles([t for t in tiles if t.cacheable], dimensions=self.dimensions)
                return tiles

            # else
        tiles = [Tile(coord) for coord in meta_tile.tiles]
        tile_collection = TileCollection(tiles)
        self.cache.load_tiles(tile_collection, dimensions=self.dimensions)
        return tile_collection


def split_meta_tiles(meta_tile: BaseImageResult, tiles: list[tuple[Optional[TileCoord], tuple[int, int]]],
                     tile_size: tuple[int, int], image_opts):
    try:
        # TODO png8
        # if not self.transparent and format == 'png':
        #     format = 'png8'
        splitter = TileSplitter(meta_tile, image_opts)
    except IOError:
        # TODO
        raise
    split_tiles = []
    for tile in tiles:
        tile_coord, crop_coord = tile
        if tile_coord is None:
            continue
        data = splitter.get_tile(crop_coord, tile_size)
        new_tile = Tile(tile_coord, cacheable=meta_tile.cacheable)
        new_tile.image_result = data
        split_tiles.append(new_tile)
    return split_tiles



from typing import TypeVar, Generic, Iterable, Iterator, Union
from urllib.parse import quote

from mapproxy.request.no_case_multi_dict import NoCaseMultiDict


V = TypeVar("V")


class RequestParams(Generic[V]):
    """
    Represents key-value request parameters with case-insensitive keys.
    Multiple values for a single key are concatenated.

    >>> qs = RequestParams(dict(foo='egg', bar='ham%eggs', baz=100)).query_string
    >>> sorted(qs.split('&'))
    ['bar=ham%25eggs', 'baz=100', 'foo=egg']
    """

    params: NoCaseMultiDict[V] = NoCaseMultiDict()

    def __init__(self, param: 'Union[dict[str, V]|NoCaseMultiDict[V]|RequestParams|None]' = None, delimiter=','):
        self.delimiter = delimiter
        if isinstance(param, RequestParams):
            self.params = param.params.copy()
        elif isinstance(param, NoCaseMultiDict) or isinstance(param, dict):
            self.params = NoCaseMultiDict(param)
        elif param is not None:
            raise ValueError('param has invalid value')

    def get(self, key, default=None):
        """
        Returns the value for `key` or the `default`.
        """
        return self.params.get(key, default)

    def set(self, key: str, value: V | Iterable[V], *, append: bool = False, unpack: bool = False) -> None:
        self.params.set(key, value, append=append, unpack=unpack)

    def update(self, mapping=(), append=False):
        """
        Update internal request parameters from an iterable of ``(key, value)``
        tuples or a dict.

        If `append` is ``True`` the value will be added to other values for
        this `key`.
        """
        self.params.update_multi(mapping, append=append)

    def __getitem__(self, key: str) -> str:
        return self.delimiter.join(map(str, self.params.get_all(key)))

    def __setitem__(self, key: str, value: V) -> None:
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        if key in self:
            del self.params[key]

    def __contains__(self, key: object) -> bool:
        return isinstance(key, str) and key in self.params

    def __getattr__(self, name: str) -> str:
        if name in self:
            return self[name]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def __iter__(self) -> Iterator[str]:
        for key, _ in self.params.items():
            yield key

    def __str__(self) -> str:
        return self.query_string

    def items(self) -> Iterator[tuple[str, V]]:
        for key, values in self.params.items_multi():
            yield key, self.delimiter.join(str(v) for v in values)

    def copy(self) -> "RequestParams[V]":
        return self.__class__(self.params, delimiter=self.delimiter)

    @property
    def query_string(self) -> str:
        """
        The map request as a query string (the order is not guaranteed).

        >>> qs = RequestParams(dict(foo='egg', bar='ham%eggs', baz=100)).query_string
        >>> sorted(qs.split('&'))
        ['bar=ham%25eggs', 'baz=100', 'foo=egg']
        """
        kv_pairs: list[str] = []

        for key, values in self.params.items_multi():
            value = self.delimiter.join(str(v) for v in values)
            encoded = quote(value.encode("utf-8"), safe=self.delimiter)
            kv_pairs.append(f"{key}={encoded}")

        return "&".join(kv_pairs)

    def with_defaults(self, defaults: "RequestParams[V]") -> "RequestParams[V]":
        """
        Return a copy with values from `defaults` applied unless explicitly set.
        """
        new = self.copy()
        for key, values in defaults.params.items_multi():
            if values != [None]:
                new.set(key, values, unpack=True)
        return new

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self.items())!r})"



# WSGI module for use with Apache mod_wsgi or gunicorn

from logging.config import fileConfig
import os


log_config = r'/mapproxy/config/logging.ini'

if os.path.isfile(log_config):
    print('Loading log config')
    fileConfig(log_config, {'here': os.path.dirname(__file__)})

multiapp_mapproxy = os.environ.get('MULTIAPP_MAPPROXY', False)

if multiapp_mapproxy:
    from mapproxy.multiapp import make_wsgi_app

    multiapp_allow_listings = os.environ.get('MULTIAPP_ALLOW_LISTINGS', False)

    print('Starting MapProxy in multi app mode')
    application = make_wsgi_app(r'/mapproxy/config/apps/', allow_listing=multiapp_allow_listings)
else:
    from mapproxy.wsgiapp import make_wsgi_app

    print('Starting MapProxy in single app mode')
    application = make_wsgi_app(r'/mapproxy/config/mapproxy.yaml', reloader=True)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from codecs import decode
import copy
import logging
import json

from functools import reduce
from io import BytesIO
from typing import Optional, Union

from lxml import etree, html

log = logging.getLogger('mapproxy.featureinfo')


class FeatureInfoDoc(object):
    content_type = None
    content: Union[None, str, bytes] = None

    def as_etree(self):
        raise NotImplementedError()

    def as_string(self):
        raise NotImplementedError()


class TextFeatureInfoDoc(FeatureInfoDoc):
    info_type = "text"

    def __init__(self, content: Union[str, bytes]):
        self.content = content

    def as_string(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return decode(self.content)  # type: ignore[arg-type]

    @classmethod
    def combine(cls, docs: list[FeatureInfoDoc]):
        result_content = [s for doc in docs if (s := doc.as_string())]

        return cls("\n".join(result_content))


class XMLFeatureInfoDoc(FeatureInfoDoc):
    info_type = "xml"
    defaultEncoding = "UTF-8"
    _etree: Optional[etree._ElementTree] = None

    def __init__(self, content: Union[str, bytes]):
        if isinstance(content, (str, bytes)):
            self.content = content
        else:
            if hasattr(content, "getroottree"):
                content = content.getroottree()
            self._etree = content
            assert hasattr(content, "getroot"), "expected etree like object"

    def as_string(self):
        if self.content is None:
            self.content = self._serialize_etree()
        if isinstance(self.content, str):
            return self.content
        else:
            return decode(self.content)  # type: ignore[arg-type]

    def as_etree(self):
        if self._etree is None:
            try:
                self._etree = self._parse_content()
            except (ValueError, etree.XMLSyntaxError):
                log.warning("Failed to parse content as XML / HTML, returning empty element tree")

        return self._etree

    def _serialize_etree(self) -> Union[str, bytes]:
        _etree = self.as_etree()
        if _etree is None:
            log.warning("Failed to serialize content as XML / HTML, returning empty string")
            return ""
        encoding = _etree.docinfo.encoding if \
            _etree.docinfo.encoding else self.defaultEncoding
        as_string = etree.tostring(_etree, encoding=encoding, xml_declaration=False)
        return decode(as_string, encoding)  # type: ignore[arg-type]

    def _get_checked_content(self, encoding: str = defaultEncoding) -> Union[str, bytes]:
        if self.content is None or not self.content:
            raise ValueError('content is None')

        content: Union[str, bytes] = self.content
        if isinstance(content, str) and not content:
            log.debug('content is empty string')
            raise ValueError('content is empty')
        if isinstance(content, str):
            return content.encode(encoding)
        return content

    def _parse_content(self):
        content = self._get_checked_content()
        if isinstance(content, str):
            return etree.ElementTree(etree.fromstring(content))
        return etree.parse(BytesIO(content))

    @classmethod
    def combine(cls, docs):
        if etree is None:
            return TextFeatureInfoDoc.combine(docs)
        doc = docs.pop(0)
        result_tree = copy.deepcopy(doc.as_etree())
        for doc in docs:
            tree = doc.as_etree()
            if tree is not None and tree.getroot() is not None:
                result_tree.getroot().extend(tree.getroot().iterchildren())
            else:
                log.debug("Document has no root element, skipping: %s", doc)

        return cls(result_tree)


class HTMLFeatureInfoDoc(XMLFeatureInfoDoc):
    info_type = "html"

    def _parse_content(self):
        content = self._get_checked_content()

        root = html.document_fromstring(content)
        return etree.ElementTree(root)

    def _serialize_etree(self):
        if self._etree is None:
            log.warning("Failed to serialize content as XML / HTML, returning empty string")
            return ""
        encoding = self._etree.docinfo.encoding if \
            self._etree.docinfo.encoding else self.defaultEncoding
        return html.tostring(self._etree.getroot(), encoding=encoding)

    @classmethod
    def combine(cls, docs):
        if etree is None:
            return TextFeatureInfoDoc.combine(docs)

        doc = docs.pop(0)
        initial_etree = doc.as_etree()
        if initial_etree is not None:
            result_tree = copy.deepcopy(initial_etree)
        else:
            result_tree = etree.ElementTree(html.fromstring('<html><body/></html>'))

        for doc in docs:
            tree = None
            try:
                tree = doc.as_etree()
            except (ValueError, etree.XMLSyntaxError):
                log.warning("Failed to parse document as XML / HTML, skipping: %s", doc.content)

            if tree is None:
                log.warning("Document has no body element, skipping: %s", doc.content)
                continue

            try:
                body = tree.getroot().body.getchildren()
            except IndexError:
                body = tree.getroot().getchildren()

            result_tree.getroot().body.extend(body)

        return cls(result_tree)


class JSONFeatureInfoDoc(FeatureInfoDoc):
    info_type = "json"

    def __init__(self, content):
        self.content = content

    def as_string(self):
        if self.content is None:
            raise ValueError('content is None')
        return self.content if isinstance(self.content, str) else decode(self.content)

    @classmethod
    def combine(cls, docs):
        contents = []
        for d in docs:
            result = json.loads(d.content)
            if result:
                contents.append(result)
        if not contents:
            return cls(json.dumps({}))
        combined = reduce(lambda a, b: merge_dict(a, b), contents)
        return cls(json.dumps(combined))


def merge_dict(base, other):
    """
    Return `base` dict with values from `conf` merged in.
    """
    for k, v in other.items():
        if k not in base:
            base[k] = v
        else:
            if isinstance(base[k], dict):
                merge_dict(base[k], v)
            elif isinstance(base[k], list):
                base[k].extend(v)
            else:
                base[k] = v
    return base


def create_featureinfo_doc(content, info_format):
    info_type = featureinfo_type(info_format)
    if info_type == "xml":
        return XMLFeatureInfoDoc(content)
    if info_type == "html":
        return HTMLFeatureInfoDoc(content)
    if info_type == "json":
        return JSONFeatureInfoDoc(content)

    return TextFeatureInfoDoc(content)


def featureinfo_type(info_format):
    info_format = info_format.split(";", 1)[
        0
    ].strip()  # remove mime options like charset
    if info_format in ("text/xml", "application/xml",
                       "application/gml+xml", "application/vnd.ogc.gml"):
        return "xml"
    if info_format == "text/html":
        return "html"
    if info_format == "application/json":
        return "json"

    return "text"


class XSLTransformer:

    def __init__(self, xsltscript, info_format=None):
        self.xsltscript = xsltscript
        self.info_type = featureinfo_type(info_format or "text/xml")

    def transform(self, input_doc):
        input_tree = input_doc.as_etree()
        xslt_tree = etree.parse(self.xsltscript)
        transform = etree.XSLT(xslt_tree)
        output_tree = transform(input_tree)
        if self.info_type == "html":
            return HTMLFeatureInfoDoc(output_tree)
        else:
            return XMLFeatureInfoDoc(output_tree)

    __call__ = transform


def combine_docs(docs, transformer=None):
    """
    Combine multiple FeatureInfoDocs.

    Combines as text, if the type of the docs differ.
    Otherwise, the type specific combine is called.

    Returns the combined document and the info_type (text, xml or json).
    info_type is None if the output is transformed, as the type is dependent on the
    transformer.
    """
    if len(set(d.info_type for d in docs)) > 1:
        # more than one info_type, combine as plain text
        doc = TextFeatureInfoDoc.combine(docs)
        infotype = "text"
    else:
        # all same type, combine with type specific handler
        infotype = docs[0].info_type
        doc = docs[0].combine(docs)

        if transformer:
            doc = transformer(doc)
            infotype = None  # defined by transformer
    return doc.as_string(), infotype



from __future__ import division

from collections import OrderedDict

from mapproxy.config.configuration.base import ConfigurationBase
from mapproxy.config.configuration.base import ConfigurationError
from mapproxy.config.spec import add_service_to_mapproxy_yaml_spec
from mapproxy.config.validator import add_service_to_config_schema
from mapproxy.service.ows import OWSServer

import logging

log = logging.getLogger('mapproxy.config')


plugin_services = {}


def register_service_configuration(service_name, service_creator,
                                   yaml_spec_service_name=None,
                                   yaml_spec_service_def=None,
                                   schema_service=None):
    """ Method used by plugins to register a new service.

        :param service_name: Name of the service
        :type service_name: str
        :param service_creator: Creator method of the service
        :type service_creator: method of type (serviceConfiguration: ServiceConfiguration, conf: dict) -> Server
        :param yaml_spec_service_name: Name of the service in the YAML configuration file
        :type yaml_spec_service_name: str
        :param yaml_spec_service_def: Definition of the service in the YAML configuration file
        :type yaml_spec_service_def: dict
        :param schema_service: JSON schema extract to insert under
        /properties/services/properties/{yaml_spec_service_name} of config-schema.json
        :type schema_service: dict
    """

    log.info('Registering configuration for plugin service %s' % service_name)
    plugin_services[service_name] = service_creator
    if yaml_spec_service_name is not None and yaml_spec_service_def is not None:
        add_service_to_mapproxy_yaml_spec(yaml_spec_service_name, yaml_spec_service_def)
    if yaml_spec_service_name is not None and schema_service is not None:
        add_service_to_config_schema(yaml_spec_service_name, schema_service)
    if yaml_spec_service_name is not None and schema_service is not None:
        add_service_to_config_schema(yaml_spec_service_name, schema_service)


class ServiceConfiguration(ConfigurationBase):
    def __init__(self, conf, context):
        if 'wms' in conf:
            if conf['wms'] is None:
                conf['wms'] = {}
            if 'md' not in conf['wms']:
                conf['wms']['md'] = {'title': 'MapProxy WMS'}

        super().__init__(conf, context)

    def services(self):
        services = []
        ows_services = []
        for service_name, service_conf in self.conf.items():
            creator = getattr(self, service_name + '_service', None)
            if not creator:
                # If not a known service, try to use the plugin mechanism
                creator = plugin_services.get(service_name, None)
                if not creator:
                    raise ValueError('unknown service: %s' % service_name)
                new_services = creator(self, service_conf or {})
            else:
                new_services = creator(service_conf or {})

            # a creator can return a list of services...
            if not isinstance(new_services, (list, tuple)):
                new_services = [new_services]

            for new_service in new_services:
                if getattr(new_service, 'service', None):
                    ows_services.append(new_service)
                else:
                    services.append(new_service)

        services.append(OWSServer(ows_services))
        return services

    def tile_layers(self, conf, use_grid_names=False):
        layers = OrderedDict()
        for layer_name, layer_conf in self.context.layers.items():
            for tile_layer in layer_conf.tile_layers(grid_name_as_path=use_grid_names):
                if not tile_layer:
                    continue
                if use_grid_names:
                    layers[tile_layer.md['name_path']] = tile_layer
                else:
                    layers[tile_layer.md['name_internal']] = tile_layer
        return layers

    def kml_service(self, conf):
        from mapproxy.service.kml import KMLServer

        md = self.context.services.conf.get('wms', {}).get('md', {}).copy()
        md.update(conf.get('md', {}))
        max_tile_age = self.context.globals.get_value('tiles.expires_hours')
        max_tile_age *= 60 * 60  # seconds
        use_grid_names = conf.get('use_grid_names', False)
        layers = self.tile_layers(conf, use_grid_names=use_grid_names)
        return KMLServer(layers, md, max_tile_age=max_tile_age, use_dimension_layers=use_grid_names)

    def tms_service(self, conf):
        from mapproxy.service.tile import TileServer

        md = self.context.services.conf.get('wms', {}).get('md', {}).copy()
        md.update(conf.get('md', {}))
        max_tile_age = self.context.globals.get_value('tiles.expires_hours')
        max_tile_age *= 60 * 60  # seconds

        origin = conf.get('origin')
        use_grid_names = conf.get('use_grid_names', False)
        layers = self.tile_layers(conf, use_grid_names=use_grid_names)
        return TileServer(layers, md, max_tile_age=max_tile_age, use_dimension_layers=use_grid_names,
                          origin=origin)

    def wmts_service(self, conf):
        from mapproxy.service.wmts import WMTSServer, WMTSRestServer

        md = self.context.services.conf.get('wms', {}).get('md', {}).copy()
        md.update(conf.get('md', {}))
        layers = self.tile_layers(conf, use_grid_names=True)

        kvp = conf.get('kvp')
        restful = conf.get('restful')

        max_tile_age = self.context.globals.get_value('tiles.expires_hours')
        max_tile_age *= 60 * 60  # seconds

        info_formats = conf.get('featureinfo_formats', [])
        info_formats = OrderedDict((f['suffix'], f['mimetype']) for f in info_formats)

        if kvp is None and restful is None:
            kvp = restful = True

        services = []
        if kvp:
            services.append(
                WMTSServer(
                    layers, md, max_tile_age=max_tile_age,
                    info_formats=info_formats,
                )
            )

        if restful:
            template = conf.get('restful_template')
            fi_template = conf.get('restful_featureinfo_template')
            if template and '{{' in template:
                # TODO remove warning in 1.6
                log.warning("double braces in WMTS restful_template are deprecated {{x}} -> {x}")
            services.append(
                WMTSRestServer(
                    layers, md, template=template,
                    fi_template=fi_template,
                    max_tile_age=max_tile_age,
                    info_formats=info_formats,
                )
            )

        return services

    def wms_service(self, conf):
        from mapproxy.service.wms import WMSServer
        from mapproxy.request.wms import Version

        md = conf.get('md', {})
        inspire_md = conf.get('inspire_md', {})
        tile_layers = self.tile_layers(conf)
        attribution = conf.get('attribution')
        strict = self.context.globals.get_value('strict', conf, global_key='wms.strict')
        on_source_errors = self.context.globals.get_value('on_source_errors',
                                                          conf, global_key='wms.on_source_errors')
        root_layer = self.context.wms_root_layer.wms_layer()
        if not root_layer:
            raise ConfigurationError("found no WMS layer")
        if not root_layer.title:
            # set title of root layer to WMS title
            root_layer.title = md.get('title')
        concurrent_layer_renderer = self.context.globals.get_value(
            'concurrent_layer_renderer', conf,
            global_key='wms.concurrent_layer_renderer')
        image_formats_names = self.context.globals.get_value('image_formats', conf,
                                                             global_key='wms.image_formats')
        image_formats = OrderedDict()
        for format in image_formats_names:
            opts = self.context.globals.image_options.image_opts({}, format)
            if opts.format in image_formats:
                log.warning('duplicate mime-type for WMS image_formats: "%s" already configured, will use last format',
                            opts.format)
            image_formats[opts.format] = opts
        info_types = conf.get('featureinfo_types')
        srs = self.context.globals.get_value('srs', conf, global_key='wms.srs')
        self.context.globals.base_config.wms.srs = srs
        srs_extents = extents_for_srs(conf.get('bbox_srs', []))

        versions = conf.get('versions')
        if versions:
            versions = sorted([Version(v) for v in versions])

        max_output_pixels = self.context.globals.get_value('max_output_pixels', conf,
                                                           global_key='wms.max_output_pixels')
        if isinstance(max_output_pixels, list):
            max_output_pixels = max_output_pixels[0] * max_output_pixels[1]

        max_tile_age = self.context.globals.get_value('tiles.expires_hours')
        max_tile_age *= 60 * 60  # seconds

        server = WMSServer(root_layer, md, attribution=attribution,
                           image_formats=image_formats, info_types=info_types,
                           srs=srs, tile_layers=tile_layers, strict=strict, on_error=on_source_errors,
                           concurrent_layer_renderer=concurrent_layer_renderer,
                           max_output_pixels=max_output_pixels, srs_extents=srs_extents,
                           max_tile_age=max_tile_age, versions=versions,
                           inspire_md=inspire_md,
                           )

        server.fi_transformers = fi_xslt_transformers(conf, self.context)

        return server

    def ogcapi_service(self, conf):
        from mapproxy.srs import SRS
        from mapproxy.service.ogcapi.server import OGCAPIServer

        root_layer = self.context.wms_root_layer.wms_layer()
        if not root_layer:
            raise ConfigurationError("found no layer")

        enable_tiles = conf.get('enable_tiles', True)
        enable_maps = conf.get('enable_maps', True)
        attribution = conf.get('attribution')
        md = conf.get('md', {})

        concurrent_layer_renderer = self.context.globals.get_value(
            'concurrent_layer_renderer', conf,
            global_key='ogcapi.concurrent_layer_renderer')

        image_formats_names = self.context.globals.get_value('image_formats', conf,
                                                             global_key='ogcapi.image_formats')
        image_formats = OrderedDict()
        for format in image_formats_names:
            opts = self.context.globals.image_options.image_opts({}, format)
            if opts.format in image_formats:
                log.warning('duplicate mime-type for OGCAPI image_formats: '
                            '"%s" already configured, will use last format',
                            opts.format)
            image_formats[opts.format] = opts

        max_output_pixels = self.context.globals.get_value('max_output_pixels', conf,
                                                           global_key='ogcapi.max_output_pixels')
        if isinstance(max_output_pixels, list):
            max_output_pixels = max_output_pixels[0] * max_output_pixels[1]

        max_tile_age = self.context.globals.get_value('tiles.expires_hours')
        max_tile_age *= 60 * 60  # seconds

        on_source_errors = self.context.globals.get_value('on_source_errors',
                                                          conf, global_key='ogcapi.on_source_errors')

        default_dataset_layers = conf.get('default_dataset_layers', None)
        if default_dataset_layers:
            layers = root_layer.child_layers()
            default_dataset_layers = [layers[id] for id in default_dataset_layers]

        grid_configs = self.context.grids

        map_srs = self.context.globals.get_value('map_srs', conf, global_key='ogcapi.map_srs')
        if map_srs:
            map_srs = [SRS(srs) for srs in map_srs]
        else:
            map_srs = []

        return OGCAPIServer(root_layer,
                            enable_tiles=enable_tiles,
                            enable_maps=enable_maps,
                            attribution=attribution,
                            metadata=md,
                            image_formats=image_formats,
                            max_tile_age=max_tile_age,
                            on_error=on_source_errors,
                            concurrent_layer_renderer=concurrent_layer_renderer,
                            max_output_pixels=max_output_pixels,
                            grid_configs=grid_configs,
                            map_srs=map_srs,
                            default_dataset_layers=default_dataset_layers)

    def demo_service(self, conf):
        from mapproxy.service.demo import DemoServer
        services = list(self.context.services.conf.keys())
        md = self.context.services.conf.get('wms', {}).get('md', {}).copy()
        md.update(conf.get('md', {}))
        layers = OrderedDict()
        for layer_name, layer_conf in self.context.layers.items():
            lyr = layer_conf.wms_layer()
            if lyr:
                layers[layer_name] = lyr
        image_formats = self.context.globals.get_value('image_formats', conf, global_key='wms.image_formats')
        srs = self.context.globals.get_value('srs', conf, global_key='wms.srs')
        tms_conf = self.context.services.conf.get('tms', {}) or {}
        use_grid_names = tms_conf.get('use_grid_names', False)
        tile_layers = self.tile_layers(tms_conf, use_grid_names=use_grid_names)

        # WMTS restful template
        wmts_conf = self.context.services.conf.get('wmts', {}) or {}
        from mapproxy.service.wmts import WMTSRestServer
        if wmts_conf:
            restful_template = wmts_conf.get('restful_template', WMTSRestServer.default_template)
        else:
            restful_template = WMTSRestServer.default_template

        if 'wmts' in self.context.services.conf:
            kvp = wmts_conf.get('kvp')
            restful = wmts_conf.get('restful')

            if kvp or kvp is None:
                services.append('wmts_kvp')
            if restful or restful is None:
                services.append('wmts_restful')

        if 'wms' in self.context.services.conf:
            versions = self.context.services.conf['wms'].get('versions', ['1.1.1'])
            if '1.1.1' in versions:
                # demo service only supports 1.1.1, use wms_111 as an indicator
                services.append('wms_111')

        layers = OrderedDict(sorted(layers.items(), key=lambda x: x[1].name))
        background = self.context.globals.get_value('background', conf)

        return DemoServer(
            layers, md, tile_layers=tile_layers, image_formats=image_formats, srs=srs, services=services,
            restful_template=restful_template, background=background)


def extents_for_srs(bbox_srs):
    from mapproxy.extent import DefaultMapExtent, MapExtent
    from mapproxy.srs import SRS
    extents = {}
    for srs in bbox_srs:
        if isinstance(srs, str):
            bbox = DefaultMapExtent()
        else:
            srs, bbox = srs['srs'], srs['bbox']
            bbox = MapExtent(bbox, SRS(srs))

        extents[srs] = bbox

    return extents


def fi_xslt_transformers(conf, context):
    from mapproxy.featureinfo import XSLTransformer
    fi_transformers = {}
    fi_xslt = conf.get('featureinfo_xslt')
    if fi_xslt:
        for info_type, fi_xslt in fi_xslt.items():
            fi_xslt = context.globals.abspath(fi_xslt)
            fi_transformers[info_type] = XSLTransformer(fi_xslt)
    return fi_transformers



from __future__ import division

from mapproxy.config.configuration.base import ConfigurationError
from mapproxy.config.configuration.base import ConfigurationBase
from mapproxy.util.py import memoize

import logging

log = logging.getLogger('mapproxy.config')


class GridConfiguration(ConfigurationBase):
    @memoize
    def tile_grid(self):
        from mapproxy.grid.tile_grid import tile_grid

        if 'base' in self.conf:
            base_grid_name = self.conf['base']
            if base_grid_name not in self.context.grids:
                raise ConfigurationError('unknown base %s for grid %s' % (base_grid_name, self.conf['name']))
            conf = self.context.grids[base_grid_name].conf.copy()
            conf.update(self.conf)
            conf.pop('base')
            self.conf = conf
        else:
            conf = self.conf
        align_with = None
        if 'align_resolutions_with' in self.conf:
            align_with_grid_name = self.conf['align_resolutions_with']
            align_with = self.context.grids[align_with_grid_name].tile_grid()

        tile_size = self.context.globals.get_value('tile_size', conf,
                                                   global_key='grid.tile_size')
        conf['tile_size'] = tuple(tile_size)
        tile_size = tuple(tile_size)

        stretch_factor = self.context.globals.get_value('stretch_factor', conf,
                                                        global_key='image.stretch_factor')
        max_shrink_factor = self.context.globals.get_value('max_shrink_factor', conf,
                                                           global_key='image.max_shrink_factor')

        if conf.get('origin') is None:
            log.warning(
                'grid %s does not have an origin. default origin will change from sw (south/west) to nw (north-west)'
                ' with MapProxy 2.0', conf['name'])

        grid = tile_grid(
            name=conf['name'],
            srs=conf.get('srs'),
            tile_size=tile_size,
            min_res=conf.get('min_res'),
            max_res=conf.get('max_res'),
            res=conf.get('res'),
            res_factor=conf.get('res_factor', 2.0),
            threshold_res=conf.get('threshold_res'),
            bbox=conf.get('bbox'),
            bbox_srs=conf.get('bbox_srs'),
            num_levels=conf.get('num_levels'),
            stretch_factor=stretch_factor,
            max_shrink_factor=max_shrink_factor,
            align_with=align_with,
            origin=conf.get('origin')
        )

        return grid



from __future__ import division

from abc import ABC, abstractmethod
from typing import Optional

from mapproxy.image import BaseImageResult
from mapproxy.grid.resolutions import ResolutionRange
from mapproxy.layer import BlankImageError
from mapproxy.image.opts import ImageOptions
from mapproxy.util.coverage import Coverage
from mapproxy.query import MapQuery


class MapLayer(ABC):
    supports_meta_tiles = False

    res_range: Optional[ResolutionRange] = None

    coverage: Optional[Coverage] = None

    def __init__(self, image_opts=None):
        self.image_opts = image_opts or ImageOptions()

    @property
    def opacity(self) -> Optional[float]:
        return self.image_opts.opacity

    @opacity.setter
    def opacity(self, value: Optional[float]):
        self.image_opts.opacity = value

    def is_opaque(self, query):
        """
        Whether the query result is opaque.

        This method is used for optimizations: layers below an opaque
        layer can be skipped. As sources with `transparent: false`
        still can return transparent images (min_res/max_res/coverages),
        implementations of this method need to be certain that the image
        is indeed opaque. is_opaque should return False if in doubt.
        """
        return False

    def check_res_range(self, query):
        if (self.res_range and
                not self.res_range.contains(query.bbox, query.size, query.srs)):
            raise BlankImageError()

    @abstractmethod
    def get_map(self, query: MapQuery) -> BaseImageResult:
        pass

    def combined_layer(self, other, query):
        return None



# This file is part of the MapProxy project.
# Copyright (C) 2016 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import calendar
import hashlib
import sys
import threading
from typing import Optional

from mapproxy.cache.tile import TileCollection
from mapproxy.cache.tile import Tile
from mapproxy.image import ImageResult
from mapproxy.cache import path
from mapproxy.cache.base import tile_buffer, TileCacheBase
from mapproxy.util import async_
from mapproxy.util.py import reraise_exception
from urllib import request as urllib2

from mapproxy.util.coverage import Coverage

try:
    import boto3
    import botocore
except ImportError:
    boto3 = None


import logging
log = logging.getLogger('mapproxy.cache.s3')


_s3_sessions_cache = threading.local()


def s3_session(profile_name=None):
    if not hasattr(_s3_sessions_cache, 'sessions'):
        _s3_sessions_cache.sessions = {}
    if profile_name not in _s3_sessions_cache.sessions:
        _s3_sessions_cache.sessions[profile_name] = boto3.session.Session(profile_name=profile_name)
    return _s3_sessions_cache.sessions[profile_name]


class S3ConnectionError(Exception):
    pass


class S3Cache(TileCacheBase):

    def __init__(self, base_path, file_ext, directory_layout='tms',
                 bucket_name='mapproxy', profile_name=None, region_name=None, endpoint_url=None,
                 _concurrent_writer=4, access_control_list=None, coverage: Optional[Coverage] = None,
                 use_http_get=False):
        super().__init__(coverage)
        md5 = hashlib.new('md5', base_path.encode('utf-8') + bucket_name.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = md5.hexdigest()
        self.bucket_name = bucket_name
        self.profile_name = profile_name
        self.region_name = region_name
        self.endpoint_url = endpoint_url
        self.access_control_list = access_control_list
        self.use_http_get = use_http_get

        try:
            self.bucket = self.conn().head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                raise S3ConnectionError('No such bucket: %s' % bucket_name)
            elif e.response['Error']['Code'] == '403':
                raise S3ConnectionError('Access denied. Check your credentials')
            else:
                raise reraise_exception(
                    S3ConnectionError('Unknown error: %s' % e),
                    sys.exc_info(),
                )

        self.base_path = base_path
        self.file_ext = file_ext
        self._concurrent_writer = _concurrent_writer

        self._tile_location, _ = path.location_funcs(layout=directory_layout)

    def get_bucket_url(self, tile):
        return f"https://{self.bucket_name}.s3.{self.region_name}.amazonaws.com/{self.tile_key(tile)}"

    def tile_key(self, tile):
        return self._tile_location(tile, self.base_path, self.file_ext).lstrip('/')

    def conn(self):
        if boto3 is None:
            raise ImportError("S3 Cache requires 'boto3' package.")

        try:
            return s3_session(self.profile_name).client(
                "s3", region_name=self.region_name, endpoint_url=self.endpoint_url)
        except Exception as e:
            raise S3ConnectionError('Error during connection %s' % e)

    def load_tile_metadata(self, tile: Tile, dimensions=None):
        if tile.timestamp:
            return
        self.is_cached(tile, dimensions=dimensions)

    def _set_metadata(self, response, tile):
        if 'LastModified' in response:
            tile.timestamp = calendar.timegm(response['LastModified'].timetuple())
        if 'ContentLength' in response:
            tile.size = response['ContentLength']

    def is_cached(self, tile: Tile, dimensions=None) -> bool:
        if tile.is_missing():
            if self.use_http_get:
                try:
                    req = urllib2.Request(self.get_bucket_url(tile))
                    response = urllib2.urlopen(req)
                    self._set_metadata(response.info(), tile)
                except urllib2.HTTPError as e:
                    if e.code == 403:
                        return False
                    raise
            else:
                key = self.tile_key(tile)
                try:
                    r = self.conn().head_object(Bucket=self.bucket_name, Key=key)
                    self._set_metadata(r, tile)
                except botocore.exceptions.ClientError as e:
                    if e.response['Error']['Code'] in ('404', 'NoSuchKey'):
                        return False
                    raise

        return True

    def load_tiles(self, tiles: TileCollection, with_metadata=True, dimensions=None) -> bool:
        p = async_.Pool(min(4, len(tiles)))
        return all(p.map(self.load_tile, tiles))

    def load_tile(self, tile: Tile, with_metadata=True, dimensions=None) -> bool:
        if not tile.is_missing():
            return True

        key = self.tile_key(tile)
        log.debug('S3:load_tile, key: %s' % key)

        if self.use_http_get:
            try:
                req = urllib2.Request(self.get_bucket_url(tile))
                tile.image_result = ImageResult(urllib2.urlopen(req))
            except urllib2.HTTPError as e:
                if e.code == 403:
                    return False
                raise
        else:
            try:
                r = self.conn().get_object(Bucket=self.bucket_name, Key=key)
                self._set_metadata(r, tile)
                tile.image_result = ImageResult(r['Body'])
            except botocore.exceptions.ClientError as e:
                # moto get_object can return Error wrapped in Errors...
                error = e.response.get('Errors', e.response)['Error']
                if error['Code'] in ('404', 'NoSuchKey'):
                    return False
                raise

        return True

    def remove_tile(self, tile, dimensions=None):
        key = self.tile_key(tile)
        log.debug('remove_tile, key: %s' % key)
        self.conn().delete_object(Bucket=self.bucket_name, Key=key)

    def store_tiles(self, tiles, dimensions=None):
        p = async_.Pool(min(self._concurrent_writer, len(tiles)))
        p.map(self.store_tile, tiles)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return

        key = self.tile_key(tile)
        log.debug('S3: store_tile, key: %s' % key)

        extra_args = {}
        if self.file_ext in ('jpeg', 'png'):
            extra_args['ContentType'] = 'image/' + self.file_ext
        if self.access_control_list:
            extra_args['ACL'] = self.access_control_list
        with tile_buffer(tile) as buf:
            self.conn().upload_fileobj(
                NopCloser(buf),  # upload_fileobj closes buf, wrap in NopCloser
                self.bucket_name,
                key,
                ExtraArgs=extra_args)


class NopCloser(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def close(self):
        pass

    def __getattr__(self, name):
        return getattr(self.wrapped, name)
